"""
A class for containing the set of strategies that
CombinatorialSpecificationSearcher (CSS) should apply when searching for a
CombinatorialSpecification.

The default setting treats the pack as follows:
    - inferral:     These srategies are applied to every comb_class first when
                    found. They should all be equivalence strategies and if a
                    strategy returns a rule, then we only work from the child.
    - initial:      After inferral we apply all of the initial strategies to
                    a comb_class. After the initial strategies are applied, the
                    the comb_class is added to the next queue, waiting for the
                    level to change (when the next becomes curr).
    - expansion:    The expansion strategies are set of sets of strategies
                    S1, ..., Sk. The strategies S1 are applied to all
                    comb_class in curr queue, then S2 to all, then S3, etc,
                    until all of the strategies are applied.
    - verification: A verification strategy is applied to every comb_class when
                    it is first found.
    - symmetry:     Symmetries are applied to a comb_class when it is first
                    found.
(the above settings can be edited by creating an alternative CSSQueue class)

If the iterative boolean is True, then CSS will search for an iterative
specification.
"""
from itertools import chain
from typing import Iterable, Iterator, Type, TypeVar

from .strategy import CSSstrategy, Strategy

PackType = TypeVar("PackType", bound="StrategyPack")

__all__ = "StrategyPack"


class StrategyPack:
    """
    A pack used by CSS to guide how to search for a specification.
    """

    def __init__(
        self,
        initial_strats: Iterable[CSSstrategy],
        inferral_strats: Iterable[CSSstrategy],
        expansion_strats: Iterable[Iterable[CSSstrategy]],
        ver_strats: Iterable[CSSstrategy],
        name: str,
        symmetries: Iterable[CSSstrategy] = None,
        iterative: bool = False,
    ):
        self.name = name
        self.initial_strats = tuple(initial_strats)
        self.inferral_strats = tuple(inferral_strats)
        self.ver_strats = tuple(ver_strats)
        self.expansion_strats = tuple(tuple(x) for x in expansion_strats)
        self.symmetries = tuple(symmetries) if symmetries is not None else tuple()
        self.iterative = iterative

    def __iter__(self) -> Iterator[CSSstrategy]:
        """Iterator over all the strategies in the pack."""
        return chain(
            self.initial_strats,
            self.ver_strats,
            self.inferral_strats,
            self.symmetries,
            *self.expansion_strats,
        )

    def __contains__(self, strategy: CSSstrategy) -> bool:
        """
        Check if the pack contains a strategy.

        Two strategy from the same Strategy class are consider the same even if
        they have different parameter.
        """
        return any(strat.__class__ == strategy.__class__ for strat in self)

    def __repr__(self) -> str:
        s = "{}(\n".format(self.__class__.__name__)
        s += "    initial_strats=[\n"
        for st in self.initial_strats:
            s += "        {!r},\n".format(st)
        s += "    ], ver_strats=[\n"
        for st in self.ver_strats:
            s += "        {!r},\n".format(st)
        s += "    ], inferral_strats=[\n"
        for st in self.inferral_strats:
            s += "        {!r},\n".format(st)
        s += "    ], expansion_strats=[\n"
        for sl in self.expansion_strats:
            s += "        [\n"
            for st in sl:
                s += "            {!r},\n".format(st)
            s += "        ],\n"
        s += "    ],\n"
        s += "    name={!r},\n".format(self.name)
        s += "    symmetries={!r},\n".format(self.symmetries)
        s += "    iterative={!r},\n".format(self.iterative)
        s += ")"
        return s

    def __str__(self) -> str:
        string = (
            "Looking for {} combinatorial specification" " with the strategies:\n"
        ).format("iterative" if self.iterative else "recursive")
        initial_strats = ", ".join(map(str, self.initial_strats))
        infer_strats = ", ".join(map(str, self.inferral_strats))
        verif_strats = ", ".join(map(str, self.ver_strats))
        string += "Inferral: {}\n".format(infer_strats)
        string += "Initial: {}\n".format(initial_strats)
        string += "Verification: {}\n".format(verif_strats)
        if self.symmetries:
            symme_strats = ", ".join(map(str, self.symmetries))
            string += "Symmetries: {}\n".format(symme_strats)
        for i, strategies in enumerate(self.expansion_strats):
            strats = ", ".join(map(str, strategies))
            string += "Set {}: {}\n".format(i + 1, strats)
        return string

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StrategyPack):
            return NotImplemented
        return self.__dict__ == other.__dict__

    # JSON methods
    def to_jsonable(self) -> dict:
        """
        Return a JSON serializable dictionary.
        """
        return {
            "name": self.name,
            "initial_strats": [s.to_jsonable() for s in self.initial_strats],
            "inferral_strats": [s.to_jsonable() for s in self.inferral_strats],
            "ver_strats": [s.to_jsonable() for s in self.ver_strats],
            "expansion_strats": [
                [s.to_jsonable() for s in strat_list]
                for strat_list in self.expansion_strats
            ],
            "symmetries": [s.to_jsonable() for s in self.symmetries],
            "iterative": self.iterative,
        }

    @classmethod
    def from_dict(cls: Type[PackType], d: dict) -> PackType:
        """
        Return a StrategyPack from the dict object return by the to_jsonable method.
        """
        name = d["name"]
        initial_strats = [
            Strategy.from_dict(strat_dict) for strat_dict in d["initial_strats"]
        ]
        inferral_strats = [
            Strategy.from_dict(strat_dict) for strat_dict in d["inferral_strats"]
        ]
        ver_strats = [Strategy.from_dict(strat_dict) for strat_dict in d["ver_strats"]]
        expansion_strats = [
            [Strategy.from_dict(strat_dict) for strat_dict in strat_list]
            for strat_list in d["expansion_strats"]
        ]
        symmetries = (
            [Strategy.from_dict(strat_dict) for strat_dict in d["symmetries"]]
            if "symmetries" in d
            else []
        )
        return cls(
            initial_strats,
            inferral_strats,
            expansion_strats,
            ver_strats,
            name,
            symmetries=symmetries,
            iterative=d.get("iterative", False),
        )

    # Method to add power to a pack
    # Pack are immutable, these methods return a new pack.
    def add_initial(
        self: PackType, strategy: CSSstrategy, name_ext: str = "", apply_first=False
    ) -> PackType:
        """
        Create a new pack with an additional initial strategy and append
        name_ext to the name of the pack.
        """
        if strategy in self:
            raise ValueError(
                ("The strategy {!r} is already in pack." "".format(strategy))
            )
        if apply_first:
            initial_strats = (strategy,) + self.initial_strats
        else:
            initial_strats = self.initial_strats + (strategy,)
        return self.__class__(
            name="_".join([self.name, name_ext]) if name_ext else self.name,
            initial_strats=initial_strats,
            ver_strats=self.ver_strats,
            inferral_strats=self.inferral_strats,
            expansion_strats=self.expansion_strats,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def add_inferral(
        self: PackType, strategy: CSSstrategy, name_ext: str = ""
    ) -> PackType:
        """
        Create a new pack with an additional inferral strategy and append
        name_ext to the name of the pack.
        """
        if strategy in self:
            raise ValueError(
                ("The strategy {!r} is already in pack." "".format(strategy))
            )
        return self.__class__(
            name="_".join([self.name, name_ext]) if name_ext else self.name,
            initial_strats=self.initial_strats,
            ver_strats=self.ver_strats,
            inferral_strats=self.inferral_strats + (strategy,),
            expansion_strats=self.expansion_strats,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def add_symmetry(
        self: PackType, strategy: CSSstrategy, name_ext: str = ""
    ) -> PackType:
        """
        Create a new pack with an additional symmetry strategy and append
        name_ext to the name of the pack.
        """
        if strategy in self:
            raise ValueError(
                ("The strategy {!r} is already in pack." "".format(strategy))
            )
        return self.__class__(
            name="_".join([self.name, name_ext]) if name_ext else self.name,
            initial_strats=self.initial_strats,
            ver_strats=self.ver_strats,
            inferral_strats=self.inferral_strats,
            expansion_strats=self.expansion_strats,
            symmetries=self.symmetries + (strategy,),
            iterative=self.iterative,
        )

    def add_verification(
        self: PackType,
        strategy: CSSstrategy,
        name_ext: str = "",
        replace: bool = False,
        apply_first: bool = False,
    ) -> PackType:
        """
        Create a new pack with an additional verification strategy and append
        name_ext to the name of the pack.
        If replace is set, it will ignore old verification strategies
        """
        if not replace and strategy in self:
            raise ValueError(
                ("The strategy {!r} is already in pack." "".format(strategy))
            )
        if apply_first:
            ver_strats = (strategy,) + (tuple() if replace else self.ver_strats)
        else:
            ver_strats = (tuple() if replace else self.ver_strats) + (strategy,)
        return self.__class__(
            name="_".join([self.name, name_ext]) if name_ext else self.name,
            initial_strats=self.initial_strats,
            ver_strats=ver_strats,
            inferral_strats=self.inferral_strats,
            expansion_strats=self.expansion_strats,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def make_iterative(self: PackType, name_ext: str = "") -> PackType:
        """
        Create a new pack with same strategies but iterative
        """
        return self.__class__(
            name="_".join([self.name, name_ext]) if name_ext else self.name,
            initial_strats=self.initial_strats,
            ver_strats=self.ver_strats,
            inferral_strats=self.inferral_strats,
            expansion_strats=self.expansion_strats,
            symmetries=self.symmetries,
            iterative=True,
        )



"""
The constructor class contains all the method, and logic, needed to get the
enumeration, generate objects, and sample objects.

The default constructors implemented are:
- CartesianProduct
- DisjointUnion
- Empty

Currently the constructors are implemented in one variable, namely 'n' which is
used throughout to denote size.
"""
import abc
from collections import defaultdict
from itertools import product
from random import randint
from typing import (
    Callable,
    Counter,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from sympy import Eq, Function

from comb_spec_searcher import utils
from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)

__all__ = ("Constructor", "CartesianProduct", "DisjointUnion")

T = TypeVar("T")


class Constructor(abc.ABC, Generic[CombinatorialClassType, CombinatorialObjectType]):
    """The constructor is akin to the 'counting function' in the comb exp paper."""

    @abc.abstractmethod
    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        """
        Return the sympy.Eq in the form lhs_func = f(rhs_funcs).
        """

    @abc.abstractmethod
    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        """
        Return the reliance profile. That is for the parameters given,
        which parameters of each individual subclass are required.
        """

    @abc.abstractmethod
    def get_terms(self, subterms: SubTerms, n: int) -> Terms:
        """
        Return the terms for n given the subterms of the children.
        """

    @abc.abstractmethod
    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        """
        Iterate over all the possibe subobjs/image of the bijection implied by
        the constructor together with the parameters.
        """

    @abc.abstractmethod
    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int
    ):
        """Return a randomly sampled subobjs/image of the bijection implied
        by the constructor."""


class CartesianProduct(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    The CartesianProduct is initialised with the children of the rule that is
    being counted. These are needed in the reliance profile. In particular,
    the CombinatorialClass that you are counting must have implemented the
    methods is_atom', 'minimum_size_of_object', 'get_minimum_values' which are
    needed to ensure that the recursions are productive.

    This CartesianProduct constructor considers compositions all of the
    parameters including n. then these should be passed to one other factor.
    The details of how the parameters map forward must be given using
    'extra_parameters'. This is a tuple, where the ith dictionary tells the
    constructor for each parent variable maps to on the child. If it does not
    map to the child it should not be a key in the dictionary.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Iterable[CombinatorialClassType],
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):

        children = tuple(children)

        if extra_parameters is not None:
            self.extra_parameters = tuple(extra_parameters)
        else:
            self.extra_parameters = tuple(dict() for _ in children)
        self._children_param_maps = self._build_children_param_map(parent, children)

        self.minimum_sizes = {"n": parent.minimum_size_of_object()}
        for k in parent.extra_parameters:
            self.minimum_sizes[k] = parent.get_minimum_value(k)

        self.min_child_sizes = tuple(
            {"n": child.minimum_size_of_object()} for child in children
        )
        self.max_child_sizes = tuple(
            {"n": child.minimum_size_of_object()} if child.is_atom() else {}
            for child in children
        )
        self.parent_parameters = ("n",) + parent.extra_parameters

        for (idx, child), parameters in zip(enumerate(children), self.extra_parameters):
            for k in parent.extra_parameters:
                self.min_child_sizes[idx][k] = (
                    child.get_minimum_value(parameters[k]) if k in parameters else 0
                )
                if k not in parameters:
                    self.max_child_sizes[idx][k] = 0
                elif child.is_atom():
                    self.max_child_sizes[idx][k] = child.get_minimum_value(
                        parameters[k]
                    )

    @property
    def min_sizes(self) -> Tuple[int, ...]:
        """
        Lower bound on the sizes of combinatorial object on the children.
        """
        return tuple(d["n"] for d in self.min_child_sizes)

    @property
    def max_sizes(self) -> Tuple[Optional[int], ...]:
        """
        Upper bounds on the size of object on the children.

        None means no upper bound.
        """
        return tuple(d.get("n", None) for d in self.max_child_sizes)

    def _build_children_param_map(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
    ) -> Tuple[ParametersMap, ...]:
        map_list: List[ParametersMap] = []
        num_parent_params = len(parent.extra_parameters)
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        for child, extra_param in zip(children, self.extra_parameters):
            reversed_extra_param: Dict[str, List[str]] = defaultdict(list)
            for parent_var, child_var in extra_param.items():
                reversed_extra_param[child_var].append(parent_var)
            child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
                tuple(
                    map(
                        parent_param_to_pos.__getitem__,
                        reversed_extra_param[child_param],
                    )
                )
                for child_param in child.extra_parameters
            )
            map_list.append(
                self._build_param_map(child_pos_to_parent_pos, num_parent_params)
            )
        return tuple(map_list)

    @staticmethod
    def _build_param_map(
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...], num_parent_params: int
    ) -> ParametersMap:
        """
        Build the ParametersMap that will map according to the given child pos to parent
        pos map given.
        """

        def param_map(param: Parameters) -> Parameters:
            new_params = [0 for _ in range(num_parent_params)]
            for pos, value in enumerate(param):
                parent_pos = child_pos_to_parent_pos[pos]
                for p in parent_pos:
                    new_params[p] += value
            return tuple(new_params)

        return param_map

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        res = 1
        for extra_parameters, rhs_func in zip(self.extra_parameters, rhs_funcs):
            res *= rhs_func.subs(
                {child: parent for parent, child in extra_parameters.items()},
                simultaneous=True,
            )
        return Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        #  TODO: consider when parameters are subsets of each other etc
        assert all(
            set(["n", *parameters]) == set(min_child_sizes)
            for min_child_sizes in self.min_child_sizes
        )
        parameters["n"] = n
        return tuple(
            {
                k: tuple(
                    range(
                        min_child_sizes[k],
                        min(
                            [
                                parameters[k]
                                - self.minimum_sizes[k]
                                + min_child_sizes[k]
                                + 1
                            ]
                            + (
                                [max_child_sizes[k] + 1] if k in max_child_sizes else []
                            ),
                        ),
                    )
                )
                for k in min_child_sizes
            }
            for min_child_sizes, max_child_sizes in zip(
                self.min_child_sizes, self.max_child_sizes
            )
        )

    def _valid_compositions(
        self, n: int, **parameters: int
    ) -> Iterator[Tuple[Dict[str, int], ...]]:
        reliance_profile = self.reliance_profile(n, **parameters)

        def _helper(
            minmaxes: Tuple[Dict[str, Tuple[int, int]], ...], **parameters: int
        ):
            if len(minmaxes) == 1:
                minmax = minmaxes[0]
                if all(
                    minmax[k][0] <= parameters[k] <= minmax[k][1]
                    for k in self.parent_parameters
                ):
                    yield ({**parameters},)
                return

            still_to_come = {
                k: sum(minmax[k][0] for minmax in minmaxes[1:])
                for k in self.parent_parameters
            }
            max_available = {
                k: sum(minmax[k][1] for minmax in minmaxes[1:])
                for k in self.parent_parameters
            }
            minmax = minmaxes[0]
            for values in product(
                *[
                    range(
                        max(minmax[k][0], parameters[k] - max_available[k]),
                        min(minmax[k][1], parameters[k] - still_to_come[k]) + 1,
                    )
                    for k in self.parent_parameters
                ]
            ):
                params = dict(zip(self.parent_parameters, values))
                update_params = {
                    k: parameters[k] - params[k] for k in self.parent_parameters
                }
                for comp in _helper(minmaxes[1:], **update_params):
                    yield (params,) + comp

        if all(all(profile.values()) for profile in reliance_profile):
            minmaxes: Tuple[Dict[str, Tuple[int, int]], ...] = tuple(
                {k: (min(profile[k]), max(profile[k])) for k in self.parent_parameters}
                for profile in reliance_profile
            )
            parameters["n"] = n
            yield from _helper(minmaxes, **parameters)

    def get_extra_parameters(
        self, child_parameters: Tuple[Dict[str, int], ...]
    ) -> Optional[List[Dict[str, int]]]:
        """
        Will return the extra parameters dictionary based on the given child
        parameters given. If there is a contradiction, that is some child
        parameter is given two or more different values that do not match,
        then None will be returned indicating that there is a contradiction,
        and so there are no objects satisfying the child parameters.
        """
        res: List[Dict[str, int]] = []
        for params, map_params in zip(child_parameters, self.extra_parameters):
            assert all(
                params[k] == 0
                for k in self.parent_parameters
                if k not in map_params and k != "n"
            )
            extra_params: Dict[str, int] = {"n": params["n"]}
            for k in map_params:
                mapped_k = map_params[k]
                if mapped_k not in extra_params:
                    extra_params[mapped_k] = params[k]
                elif extra_params[mapped_k] != params[k]:
                    return None
            res.append(extra_params)
        return res

    def get_terms(self, subterms: SubTerms, n: int) -> Terms:
        new_terms: Terms = Counter()
        size_compositions = utils.compositions(
            n, len(subterms), self.min_sizes, self.max_sizes
        )
        for sizes in size_compositions:
            for param_value_pairs in self._params_value_pairs_combinations(
                sizes, subterms
            ):
                new_param = self._new_param(*(p for p, _ in param_value_pairs))
                new_terms[new_param] += utils.prod((v for _, v in param_value_pairs))
        return new_terms

    def _new_param(self, *children_params: Parameters) -> Parameters:
        """
        Computes the parameter values on the parent that the given children parameters
        contribute to.
        """
        mapped_params = (
            pmap(p) for pmap, p in zip(self._children_param_maps, children_params)
        )
        return tuple(sum(vals) for vals in zip(*mapped_params))

    @staticmethod
    def _params_value_pairs_combinations(
        sizes: Tuple[int, ...],
        sub_getters: Tuple[Callable[[int], Dict[Parameters, T]], ...],
    ) -> Iterator[Tuple[Tuple[Parameters, T], ...]]:
        """"""
        children_values = (sg(s) for sg, s in zip(sub_getters, sizes))
        return product(*(c.items() for c in children_values))

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        size_compositions = utils.compositions(
            n, len(subobjs), self.min_sizes, self.max_sizes
        )
        for sizes in size_compositions:
            for param_objs_pairs in self._params_value_pairs_combinations(
                sizes, subobjs
            ):
                new_param = self._new_param(*(p for p, _ in param_objs_pairs))
                children_objs = cast(
                    Iterator[List[Optional[CombinatorialObjectType]]],
                    (objs for _, objs in param_objs_pairs),
                )
                yield (new_param, tuple(children_objs))

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int
    ):
        random_choice = randint(1, parent_count)
        total = 0
        for child_parameters in self._valid_compositions(n, **parameters):
            tmp = 1
            extra_parameters = self.get_extra_parameters(child_parameters)
            if extra_parameters is None:
                continue
            for rec, extra_params in zip(subrecs, extra_parameters):
                tmp *= rec(n=extra_params.pop("n"), **extra_params)
                if tmp == 0:
                    break
            total += tmp
            if random_choice <= total:
                extra_parameters = self.get_extra_parameters(child_parameters)
                assert extra_parameters is not None
                return tuple(
                    subsampler(n=extra_params.pop("n"), **extra_params)
                    for subsampler, extra_params in zip(subsamplers, extra_parameters)
                )

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "x"

    def __str__(self) -> str:
        return "Cartesian product"


class DisjointUnion(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    The DisjointUnion constructor takes as input the children. Each constructor
    is unique up to the length of the children being used to count.

    Extra parameters are passed on using the parameters dictionaries. Each
    dictionary's keys should be the extra variable of the child pointing the
    variable on the parent it came from.
    If a parents variable does not map to a child, then this variable must be 0
    as the child contains no occurences.

    The fixed value dictionaries passed will be used ensure that the parameter
    of a child must take on the given value.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
        fixed_values: Optional[Tuple[Dict[str, int], ...]] = None,
    ):
        self.number_of_children = len(children)
        if extra_parameters is not None:
            self.extra_parameters = extra_parameters
            assert len(extra_parameters) == len(children)
        else:
            assert not parent.extra_parameters
            self.extra_parameters = tuple(
                dict() for _ in range(self.number_of_children)
            )
        self._children_param_maps = self._build_children_param_maps(parent, children)
        self.zeroes = tuple(
            frozenset(parent.extra_parameters) - frozenset(parameter.keys())
            for parameter in self.extra_parameters
        )
        if fixed_values is not None:
            self.fixed_values = fixed_values
            assert len(fixed_values) == len(children)
        else:
            self.fixed_values = tuple({} for _ in children)

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        res = 0
        for rhs_func, extra_parameters in zip(rhs_funcs, self.extra_parameters):
            res += rhs_func.subs(
                {child: parent for parent, child in extra_parameters.items()},
                simultaneous=True,
            )
        return Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        # TODO: implement in multiple variables and use in get_recurrence
        assert not parameters, "only implemented in one variable, namely 'n'"
        return tuple({"n": (n,)} for _ in range(self.number_of_children))

    def get_extra_parameters(
        self, n: int, **parameters: int
    ) -> List[Optional[Dict[str, int]]]:
        """
        Will return the extra parameters dictionary based on the parent's
        parameters. If there is a contradiction, that is some child parameter
        is given two or more different values that do not match, then None
        will be returned for that child, indicating that 0 objects on the
        child match the parents parameters.
        """
        res: List[Optional[Dict[str, int]]] = []
        for i, extra_parameters in enumerate(self.extra_parameters):
            update_params: Dict[str, int] = {**self.fixed_values[i]}
            for parent_var, child_var in extra_parameters.items():
                updated_value = parameters[parent_var]
                if child_var not in update_params:
                    update_params[child_var] = updated_value
                elif update_params[child_var] != updated_value:
                    break
            else:
                res.append(update_params)
                continue
            res.append(None)
        return res

    def get_terms(self, subterms: SubTerms, n: int) -> Terms:
        new_terms: Terms = Counter()
        for child_terms, param_map in zip(subterms, self._children_param_maps):
            for param, value in child_terms(n).items():
                new_terms[param_map(param)] += value
        return new_terms

    @staticmethod
    def _build_param_map(
        child_pos_to_parent_pos: Tuple[Optional[Tuple[int, ...]], ...],
        num_parent_params: int,
    ) -> ParametersMap:
        """
        Build the ParametersMap that will map according to the given child pos to parent
        pos map given.

        If a child pos maps to None, then the value of the parameter must be 0 at all
        times.
        """

        def param_map(param: Parameters) -> Parameters:
            new_params = [0 for _ in range(num_parent_params)]
            for pos, value in enumerate(param):
                parent_pos = child_pos_to_parent_pos[pos]
                if parent_pos is None:
                    assert value == 0
                    continue
                for p in parent_pos:
                    assert new_params[p] == 0
                    new_params[p] = value
            return tuple(new_params)

        return param_map

    def _build_children_param_maps(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
    ) -> Tuple[ParametersMap, ...]:
        map_list: List[ParametersMap] = []
        num_parent_params = len(parent.extra_parameters)
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        for child, extra_param in zip(children, self.extra_parameters):
            reversed_extra_param: Dict[str, List[str]] = defaultdict(list)
            for parent_var, child_var in extra_param.items():
                reversed_extra_param[child_var].append(parent_var)
            child_pos_to_parent_pos: Tuple[Optional[Tuple[int, ...]], ...] = tuple(
                tuple(
                    map(
                        parent_param_to_pos.__getitem__,
                        reversed_extra_param[child_param],
                    )
                )
                if child_param in reversed_extra_param
                else None
                for child_param in child.extra_parameters
            )
            map_list.append(
                self._build_param_map(child_pos_to_parent_pos, num_parent_params)
            )
        return tuple(map_list)

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        res: List[List[Optional[CombinatorialObjectType]]]
        res = [[None] for _ in range(self.number_of_children)]
        for i, subobj in enumerate(subobjs):
            param_map = self._children_param_maps[i]
            for param, comb_objs in subobj(n).items():
                res[i] = comb_objs
                yield (param_map(param), tuple(res))
            res[i] = [None]

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int
    ):
        random_choice = randint(1, parent_count)
        total = 0
        for (idx, rec), subsampler, extra_params in zip(
            enumerate(subrecs), subsamplers, self.get_extra_parameters(n, **parameters)
        ):
            # if a parent parameter is not mapped to by some child parameter
            # then it is assumed that the value of the parent parameter must be 0
            if extra_params is None or any(
                val != 0 and k in self.zeroes[idx] for k, val in parameters.items()
            ):
                continue
            total += rec(n=n, **extra_params)
            if random_choice <= total:
                obj = subsampler(n=n, **extra_params)
                return (
                    tuple(None for _ in range(idx))
                    + (obj,)
                    + tuple(None for _ in range(len(subrecs) - idx - 1))
                )

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "+"

    def __str__(self):
        return "disjoint union"



"""
    An alternative ruledb that permits restricting the number of applications of a given
    set of strategies to a given limit.
"""
from copy import deepcopy
from itertools import combinations
from typing import Iterable, Set, Type

from logzero import logger

from comb_spec_searcher.exception import SpecificationNotFound
from comb_spec_searcher.rule_db import RuleDB
from comb_spec_searcher.strategies import AbstractStrategy
from comb_spec_searcher.tree_searcher import (
    Node,
    iterative_proof_tree_finder,
    iterative_prune,
    prune,
    smallish_random_proof_tree,
)
from comb_spec_searcher.typing import RuleKey


class LimitedStrategyRuleDB(RuleDB):
    """
    An alternative ruledb that permits restricting the number of applications of a
    given set of strategies to a given limit. The limit is a single limit for the
    total number of strategies whose type is in the given set; it is not a per-type
    limit.
    """

    def __init__(
        self,
        strategies_to_limit: Iterable[Type[AbstractStrategy]],
        limit: int,
        mark_verified: bool,
    ) -> None:
        self.strategies_to_limit = set(strategies_to_limit)
        self.limit = limit
        self.mark_verified = mark_verified
        super().__init__()

    def get_rules_up_to_equiv_using_strategies(self) -> Set[RuleKey]:
        """
        returns a set of rules, up to equivalence, that come from
        <self.strategies_to_limit>
        """
        eqv_rules_using_strategies: Set[RuleKey] = set()
        for label_rule in self:
            rule_strat = self.rule_to_strategy[label_rule]
            start, ends = label_rule
            if isinstance(rule_strat, tuple(self.strategies_to_limit)):
                temp_start = self.equivdb[start]
                temp_ends = tuple(sorted(map(self.equivdb.__getitem__, ends)))
                eqv_rules_using_strategies.add((temp_start, temp_ends))
        return eqv_rules_using_strategies

    def find_specification(
        self, label: int, minimization_time_limit: float, iterative: bool = False
    ) -> Node:
        """Search for a specification based on current data found."""
        rules_dict = self.rules_up_to_equivalence()
        # Prune all unverified labels (recursively)
        # Afterward, only verified labels remain in rules_dict. In particular, there
        # is a specification if <label> is in the rules_dict
        if iterative:
            rules_dict = iterative_prune(rules_dict, root=label)
        else:
            prune(rules_dict)  # this function removes rules not in a specification.

        # if the root is not in the pruned dict, then there's nothing to be done
        if self.equivdb[label] not in rules_dict:
            raise SpecificationNotFound("No specification for label {}".format(label))

        # find the rules that we want to limit, and consider only those in a spec
        rules_to_isolate = self.get_rules_up_to_equiv_using_strategies()
        rules_to_isolate = set(
            (lhs, rhs)
            for (lhs, rhs) in rules_to_isolate
            if lhs in rules_dict and rhs in rules_dict[lhs]
        )

        logger.debug("Found %s rules to isolate.", len(rules_to_isolate))

        if len(rules_to_isolate) < self.limit:
            rule_combos = [tuple(rules_to_isolate)]
        else:
            rule_combos = list(combinations(rules_to_isolate, self.limit))
        num_combos = len(rule_combos)
        logger.debug("%s combinations to check", num_combos)

        # for each combination of <self.limit> rules, remove all but those and
        # reprune to see if a spec still exists
        for i, combo in enumerate(rule_combos):
            logger.debug("[ %s / %s ]", i + 1, num_combos)
            temp_rules_dict = deepcopy(rules_dict)
            for lhs, rhs in rules_to_isolate:
                if (lhs, rhs) in combo:
                    continue
                try:
                    temp_rules_dict[lhs].remove(rhs)
                except KeyError:
                    pass
                if len(temp_rules_dict[lhs]) == 0:
                    del temp_rules_dict[lhs]
            prune(temp_rules_dict)

            if self.mark_verified:
                for ver_label in temp_rules_dict.keys():
                    self.set_verified(ver_label)

            if self.equivdb[label] in temp_rules_dict:
                for label_in_spec in temp_rules_dict.keys():
                    self.set_verified(label_in_spec)
                if iterative:
                    specification = iterative_proof_tree_finder(
                        temp_rules_dict, root=self.equivdb[label]
                    )
                else:
                    specification = smallish_random_proof_tree(
                        temp_rules_dict, self.equivdb[label], minimization_time_limit
                    )
                return specification

        raise SpecificationNotFound("No specification for label {}".format(label))



"""
The Strategy class is essentially following the mantra of 'strategy' from the
combinatorial explanation paper.
(https://permutatriangle.github.io/papers/2019-02-27-combex.html)

In order to use CombinatorialSpecificationSearcher, you must implement a
Strategy. The functions required are:
    - decomposition_function:   Given a combinatorial class, the decomposition
                                function should return the tuple of
                                combinatorial classes it can be counted by. The
                                function should return None if it doesn't
                                apply.
    - constructor:              Return a Constructor class. If you wish to use
                                CartesianProduct or DisjointUnion, consider
                                using the CartesianProductStrategy or
                                DisjointUnionStrategy subclasses.
    - formal_step:              A short string explaining what was done.
    - backward_map:             This is the backward mapping of the underlying
                                bijection of the strategy. If you want to
                                generate objects, or sample you must implement
                                this. You can instead raise NotImplementedError
                                if you don't wish to use these features.
    - forward_map:              This is the forward mapping of the underlying
                                bijection. See the discussion for forward map.
    - __repr__ and __str__:     This is mostly for printing purposes!
    - from_dict:                A method that can recreate the class. The dict
                                passed is empty. If your strategy needs extra
                                parameters to recreate you should overwrite the
                                to_jsonable method.
Also included in this file is a StrategyFactory class. This is useful if you
are defining a family of strategies. When the __call__ method is applied to it,
it should return an iterator of Strategy to try and apply to the comb_class.

For a VerificationStrategy you must implement the methods:
    - verified:                 Return True if the combinatorial class is
                                verified by the strategy.
    - pack                      The pack is used to count and generate the
                                objects. If the strategy doesn't have a CSS
                                strategy pack that can be used to enumerate
                                verified combinatorial classes, then you need
                                to implement the methods get_terms,
                                get_objects, and get_genf.
    - __repr__ and __str__:     This is mostly for printing purposes!
    - from_dict:                A method that can recreate the class. The dict
                                passed is empty. If your strategy needs extra
                                parameters to recreate you should overwrite the
                                to_jsonable method.
If your verification strategy is for the atoms, consider using the
AtomStrategy, relying on CombinatorialClass methods.
"""
import abc
from collections import defaultdict
from importlib import import_module
from typing import (
    TYPE_CHECKING,
    Counter,
    Dict,
    Generic,
    Iterator,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from sympy import Expr, Function, Integer, var

from comb_spec_searcher.typing import CSSstrategy, Objects, Terms

from ..combinatorial_class import (
    CombinatorialClass,
    CombinatorialClassType,
    CombinatorialObject,
    CombinatorialObjectType,
)
from ..exception import InvalidOperationError, ObjectMappingError, StrategyDoesNotApply
from .constructor import CartesianProduct, Constructor, DisjointUnion
from .rule import AbstractRule, Rule, VerificationRule

if TYPE_CHECKING:
    from comb_spec_searcher import CombinatorialSpecification

    from .strategy_pack import StrategyPack

__all__ = (
    "AbstractStrategy",
    "CartesianProductStrategy",
    "DisjointUnionStrategy",
    "Strategy",
    "StrategyFactory",
    "SymmetryStrategy",
    "VerificationStrategy",
)


def strategy_from_dict(d) -> CSSstrategy:
    """
    Return the AbstractStrategy or StrategyFactory from the json representation.
    """
    module = import_module(d.pop("class_module"))
    StratClass: Type[CSSstrategy] = getattr(module, d.pop("strategy_class"))
    assert issubclass(
        StratClass, (AbstractStrategy, StrategyFactory)
    ), "Not a valid strategy"
    return StratClass.from_dict(d)


class AbstractStrategy(
    abc.ABC, Generic[CombinatorialClassType, CombinatorialObjectType]
):
    """
    A base class for strategies for methods that Strategy and
    VerificationStrategy have in common.
    """

    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        self._ignore_parent = ignore_parent
        self._inferrable = inferrable
        self._possibly_empty = possibly_empty
        self._workable = workable

    @abc.abstractmethod
    def __call__(
        self,
        comb_class: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...] = None,
        **kwargs,
    ) -> AbstractRule[CombinatorialClassType, CombinatorialObjectType]:
        """
        Return the rule formed by using the strategy.
        """

    @property
    def ignore_parent(self) -> bool:
        """
        Return True if it is not worth expanding the parent/comb_class if
        the strategy applies.
        """
        return self._ignore_parent

    @property
    def inferrable(self) -> bool:
        """
        Return True if the children could change using inferral strategies.
        """
        return self._inferrable

    @property
    def possibly_empty(self) -> bool:
        """
        Return True if it is possible that a child is empty.
        """
        return self._possibly_empty

    @property
    def workable(self) -> bool:
        """
        Return True if the children can expanded using other strategies.
        """
        return self._workable

    @abc.abstractmethod
    def can_be_equivalent(self) -> bool:
        """
        Return True if every Rule returned with one non-empty child is an
        equivalence rule.
        """

    @abc.abstractmethod
    def decomposition_function(
        self, comb_class: CombinatorialClassType
    ) -> Optional[Tuple[CombinatorialClassType, ...]]:
        """
        Return the children of the strategy for the given comb_class. It
        should return None if it does not apply.
        """

    @abc.abstractmethod
    def formal_step(self) -> str:
        """
        Return a short string to explain what the strategy has done.
        """

    @staticmethod
    def get_eq_symbol() -> str:
        """
        Return a choice for '=' in the pretty print a '=' b '?' c of rules.
        Your choice should be a single charachter.
        """
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        """
        Return a choice for '?' in the pretty print a '=' b '?' c of rules.
        Your choice should be a single charachter.
        """
        return "?"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractStrategy):
            return NotImplemented
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        return (
            self.__class__.__name__
            + f"(ignore_parent={self.ignore_parent}, inferrable={self.inferrable},"
            f" possibly_empty={self.possibly_empty}, workable={self.workable})"
        )

    def __str__(self) -> str:
        return self.formal_step()

    def to_jsonable(self) -> dict:
        """
        Return a dictionary form of the strategy.
        """
        c = self.__class__
        return {
            "class_module": c.__module__,
            "strategy_class": c.__name__,
            "ignore_parent": self._ignore_parent,
            "inferrable": self._inferrable,
            "possibly_empty": self._possibly_empty,
            "workable": self._workable,
        }

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, d: dict) -> CSSstrategy:
        return strategy_from_dict(d)


class Strategy(AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]):
    """
    The Strategy class is essentially following the mantra of 'strategy' from the
    combinatorial explanation paper.
    (https://permutatriangle.github.io/papers/2019-02-27-combex.html)

    In order to use CombinatorialSpecificationSearcher, you must implement a
    Strategy. The functions required are:
        - decomposition_function:   Given a combinatorial class, the decomposition
                                    function should return the tuple of
                                    combinatorial classes it can be counted by. The
                                    function should return None if it doesn't
                                    apply.
        - constructor:              Return a Constructor class. If you wish to use
                                    CartesianProduct or DisjointUnion, consider
                                    using the CartesianProductStrategy or
                                    DisjointUnionStrategy subclasses.
        - formal_step:              A short string explaining what was done.
        - backward_map:             This is the backward mapping of the underlying
                                    bijection of the strategy. If you want to
                                    generate objects, or sample you must implement
                                    this. You can instead raise NotImplementedError
                                    if you don't wish to use these features.
        - forward_map:              This is the forward mapping of the underlying
                                    bijection. See the discussion for forward map.
        - __repr__ and __str__:     This is mostly for printing purposes!
        - from_dict:                A method that can recreate the class. The dict
                                    passed is empty. If your strategy needs extra
                                    parameters to recreate you should overwrite the
                                    to_jsonable method.
    """

    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=inferrable,
            possibly_empty=possibly_empty,
            workable=workable,
        )

    def __call__(
        self,
        comb_class: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...] = None,
        **kwargs,
    ) -> Rule[CombinatorialClassType, CombinatorialObjectType]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        return Rule(self, comb_class, children=children)

    @abc.abstractmethod
    def constructor(
        self,
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Constructor:
        """
        This is where the details of the 'reliance profile' and 'counting'
        functions are hidden.
        """
        if children is None:
            children = self.decomposition_function(comb_class)

    @abc.abstractmethod
    def backward_map(
        self,
        comb_class: CombinatorialClassType,
        objs: Tuple[Optional[CombinatorialObjectType], ...],
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Iterator[CombinatorialObjectType]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)

    @abc.abstractmethod
    def forward_map(
        self,
        comb_class: CombinatorialClassType,
        obj: CombinatorialObjectType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)

    def extra_parameters(
        self,
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        """
        This should be a tuple of dictionaries where the parent parameters point
        to the corresponding child parameter. Any parent parameter not
        corresponding to a child parameter must have no objects that are on
        that child.
        """
        assert not comb_class.extra_parameters, (
            "you need to update the 'extra_parameters' method in the strategy {} "
            "in order to enumerate class with multiple extra_parameters".format(
                str(self)
            )
        )
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        return tuple(dict() for _ in children)


class CartesianProductStrategy(
    Strategy[CombinatorialClassType, CombinatorialObjectType]
):
    """
    The CartesianProductStrategy is a subclass of strategy. The constructor is
    CartesianProduct. Such strategies by default assume
    ignore_parent=True, inferrable=False, possibly_empty=False, and
    workable=True.

    The bijection maps an object a -> (b1, ..., bk) where bi is the object in
    the child at index i returned by the decomposition function.
    """

    def __init__(
        self,
        ignore_parent: bool = True,
        inferrable: bool = False,
        possibly_empty: bool = False,
        workable: bool = True,
    ):
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=inferrable,
            possibly_empty=possibly_empty,
            workable=workable,
        )

    @staticmethod
    def can_be_equivalent() -> bool:
        return True

    def constructor(
        self,
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Constructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        return CartesianProduct(
            comb_class,
            children,
            extra_parameters=self.extra_parameters(comb_class, children),
        )

    @staticmethod
    def get_op_symbol() -> str:
        """
        Return a choice for 'x' in the pretty print a '=' b 'x' c of rules.
        Your choice should be a single charachter.
        """
        return "x"


class DisjointUnionStrategy(Strategy[CombinatorialClassType, CombinatorialObjectType]):
    """
    The DisjointUnionStrategy is a subclass of Strategy. The constructor used
    is DisjointUnion.

    The bijection maps an object a -> (None, ..., b, ..., None) where b is at
    the index of the child it belongs to.
    """

    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=inferrable,
            possibly_empty=possibly_empty,
            workable=workable,
        )

    @staticmethod
    def can_be_equivalent() -> bool:
        return True

    def constructor(
        self,
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> DisjointUnion:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        return DisjointUnion(
            comb_class,
            children,
            extra_parameters=self.extra_parameters(comb_class, children),
        )

    @staticmethod
    def backward_map_index(objs: Tuple[Optional[CombinatorialObjectType], ...]) -> int:
        """
        Return the index of the comb_class that the sub_object returned.
        """
        for idx, obj in enumerate(objs):
            if obj is not None:
                return idx
        raise ObjectMappingError(
            "For a disjoint union strategy, an object O is mapped to the tuple"
            "with entries being None, except at the index of the child which "
            "contains O, where it should be O."
        )

    def backward_map(
        self,
        comb_class: CombinatorialClassType,
        objs: Tuple[Optional[CombinatorialObjectType], ...],
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ) -> Iterator[CombinatorialObjectType]:
        """
        This method will enable us to generate objects, and sample.
        If it is a direct bijection, the below implementation will work!
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        idx = DisjointUnionStrategy.backward_map_index(objs)
        yield cast(CombinatorialObjectType, objs[idx])

    @staticmethod
    def get_op_symbol() -> str:
        """
        Return a choice for '+' in the pretty print a '=' b '+' c of rules.
        Your choice should be a single charachter.
        """
        return "+"


class SymmetryStrategy(
    DisjointUnionStrategy[CombinatorialClassType, CombinatorialObjectType]
):
    """General representation for a symmetry strategy."""

    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = False,
        possibly_empty: bool = False,
        workable: bool = False,
    ):
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=inferrable,
            possibly_empty=possibly_empty,
            workable=workable,
        )


class VerificationStrategy(
    AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]
):
    """
    For a VerificationStrategy you must implement the methods:
        - verified:                 Return True if the combinatorial class is
                                    verified by the strategy.
        - pack                      The pack is used to count and generate the
                                    objects. If the strategy doesn't have a CSS
                                    strategy pack that can be used to enumerate
                                    verified combinatorial classes, then you need
                                    to implement the methods get_terms,
                                    get_objects, and get_genf.
        - __repr__ and __str__:     This is mostly for printing purposes!
        - from_dict:                A method that can recreate the class. The dict
                                    passed is empty. If your strategy needs extra
                                    parameters to recreate you should overwrite the
                                    to_jsonable method.
    If your verification strategy is for the atoms, consider using the
    AtomStrategy, relying on CombinatorialClass methods.
    """

    def __init__(self, ignore_parent: bool = True):
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=False,
            possibly_empty=False,
            workable=False,
        )

    def __call__(
        self,
        comb_class: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...] = None,
        **kwargs,
    ) -> VerificationRule[CombinatorialClassType, CombinatorialObjectType]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("The combinatorial class is not verified")
        return VerificationRule(self, comb_class, children)

    @staticmethod
    def can_be_equivalent() -> bool:
        return False

    def pack(self, comb_class: CombinatorialClassType) -> "StrategyPack":
        """
        Returns a StrategyPack that finds a proof tree for the comb_class in
        which the verification strategies used are "simpler".

        The pack is assumed to produce a finite universe.
        """
        raise InvalidOperationError(f"can't find specification for {self}")

    @abc.abstractmethod
    def verified(self, comb_class: CombinatorialClassType) -> bool:
        """
        Returns True if enumeration strategy works for the combinatorial class.
        """

    def get_specification(
        self, comb_class: CombinatorialClassType
    ) -> "CombinatorialSpecification[CombinatorialClassType, CombinatorialObjectType]":
        """
        Returns a combinatorial specification for the combinatorial class.
        Raises an `StrategyDoesNotApply` if no specification can be found,
        e.g. if it is not verified.
        """
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("The combinatorial class is not verified")
        # pylint: disable=import-outside-toplevel
        from ..comb_spec_searcher import CombinatorialSpecificationSearcher

        searcher = CombinatorialSpecificationSearcher(comb_class, self.pack(comb_class))
        specification = searcher.auto_search()
        assert specification is not None, StrategyDoesNotApply(
            "Cannot find a specification"
        )
        return specification

    def get_genf(
        self,
        comb_class: CombinatorialClassType,
        funcs: Optional[Dict[CombinatorialClassType, Function]] = None,
    ) -> Expr:
        """
        Returns the generating function for the combinatorial class.
        Raises an StrategyDoesNotApply if the combinatorial class is not verified.
        """
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("The combinatorial class is not verified")
        return self.get_specification(comb_class).get_genf()

    def decomposition_function(
        self, comb_class: CombinatorialClassType
    ) -> Union[Tuple[CombinatorialClassType, ...], None]:
        """
        A combinatorial class C is marked as verified by returning a rule
        C -> (). This ensures that C is in a combinatorial specification as it
        appears exactly once on the left hand side.

        The function returns None if the verification strategy doesn't apply.
        """
        if self.verified(comb_class):
            return tuple()
        return None

    def get_terms(self, comb_class: CombinatorialClassType, n: int) -> Terms:
        """
        Return the terms for n given the subterms of the children.
        """
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("The combinatorial class is not verified")
        return self.get_specification(comb_class).get_terms(n)

    def get_objects(self, comb_class: CombinatorialClassType, n: int) -> Objects:
        """
        A method to get all the objects of a given size n partitioned by parameters.
        Raises an StrategyDoesNotApply if the combinatorial class is not verified.
        """
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("The combinatorial class is not verified")
        return self.get_specification(comb_class).get_objects(n)

    def random_sample_object_of_size(
        self, comb_class: CombinatorialClassType, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        """
        A method to sample uniformly at random from a verified combinatorial class.
        Raises an StrategyDoesNotApply if the combinatorial class is not verified.
        """
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("The combinatorial class is not verified")
        return self.get_specification(comb_class).random_sample_object_of_size(
            n, **parameters
        )

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        return d


class AtomStrategy(VerificationStrategy[CombinatorialClass, CombinatorialObject]):
    """
    A subclass for when a combinatorial class is an atom - meaning consisting
    of a single object.
    """

    def __init__(self):
        super().__init__(ignore_parent=True)

    @staticmethod
    def get_terms(comb_class: CombinatorialClass, n: int) -> Terms:
        if comb_class.extra_parameters:
            raise NotImplementedError
        if n == comb_class.minimum_size_of_object():
            return Counter([tuple()])
        return Counter()

    @staticmethod
    def get_objects(comb_class: CombinatorialClass, n: int) -> Objects:
        if comb_class.extra_parameters:
            raise NotImplementedError
        res: Objects = defaultdict(list)
        if n == comb_class.minimum_size_of_object():
            res[tuple()].append(next(comb_class.objects_of_size(n)))
        return res

    def get_genf(
        self,
        comb_class: CombinatorialClass,
        funcs: Optional[Dict[CombinatorialClass, Function]] = None,
    ) -> Expr:
        if comb_class.extra_parameters:
            raise NotImplementedError
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("Can't find generating functon for non-atom.")
        x = var("x")
        return x ** comb_class.minimum_size_of_object()

    @staticmethod
    def random_sample_object_of_size(
        comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> CombinatorialObject:
        if comb_class.extra_parameters:
            raise NotImplementedError
        if n == comb_class.minimum_size_of_object():
            obj: CombinatorialObject = next(comb_class.objects_of_size(n))
            return obj

    @staticmethod
    def verified(comb_class: CombinatorialClass) -> bool:
        return bool(comb_class.is_atom())

    @staticmethod
    def formal_step() -> str:
        return "is atom"

    @staticmethod
    def pack(comb_class: CombinatorialClass) -> "StrategyPack":
        raise InvalidOperationError("No pack for the empty strategy.")

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AtomStrategy":
        assert not d
        return cls()

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"(ignore_parent={self.ignore_parent})"

    def __str__(self) -> str:
        return "verify atoms"


class EmptyStrategy(VerificationStrategy[CombinatorialClass, CombinatorialObject]):
    """
    A subclass for when a combinatorial class is equal to the empty set.
    """

    def __init__(self):
        super().__init__(ignore_parent=True)

    @staticmethod
    def get_terms(comb_class: CombinatorialClass, n: int) -> Terms:
        return Counter()

    @staticmethod
    def get_objects(comb_class: CombinatorialClass, n: int) -> Objects:
        return defaultdict(list)

    def get_genf(
        self,
        comb_class: CombinatorialClass,
        funcs: Optional[Dict[CombinatorialClass, Function]] = None,
    ) -> Integer:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply(
                "can't find generating functon for non-empty class."
            )
        return Integer(0)

    @staticmethod
    def random_sample_object_of_size(
        comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> CombinatorialObject:
        raise StrategyDoesNotApply("Can't sample from empty set.")

    @staticmethod
    def verified(comb_class: CombinatorialClass) -> bool:
        return bool(comb_class.is_empty())

    @staticmethod
    def formal_step() -> str:
        return "is empty"

    @staticmethod
    def pack(comb_class: CombinatorialClass) -> "StrategyPack":
        raise InvalidOperationError("No pack for the empty strategy.")

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "EmptyStrategy":
        assert not d
        return cls()

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return "the empty strategy"


class StrategyFactory(abc.ABC, Generic[CombinatorialClassType]):
    """
    The StrategyFactory class can be used instead of the Strategy class if
    you wish to expand a combinatorial class with a family of strategies.
    """

    @abc.abstractmethod
    def __call__(
        self, comb_class: CombinatorialClassType, **kwargs
    ) -> Iterator[Union[AbstractRule, AbstractStrategy]]:
        """
        Returns the results of the strategy on a comb_class.
        """

    @abc.abstractmethod
    def __str__(self) -> str:
        """
        Return the name of the strategy.
        """

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    def __eq__(self, other: object) -> bool:
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        """
        Hash function for the strategy.

        As we don't expect a use case were many object for the same class
        strategy are used. This hash function should perform correctly.

        # TODO: do better, why is it hashable at all?
        """
        return hash(self.__class__)

    def to_jsonable(self) -> dict:
        """
        Return a dictionary form of the strategy.
        """
        c = self.__class__
        return {
            "class_module": c.__module__,
            "strategy_class": c.__name__,
        }

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, d: dict) -> CSSstrategy:
        """
        Return the strategy from the json representation.
        """
        return strategy_from_dict(d)



"""
A database for rules.
"""
import itertools
from typing import Iterable, Iterator, MutableMapping, Set, Tuple, Union, cast

from comb_spec_searcher.class_db import ClassDB
from comb_spec_searcher.equiv_db import EquivalenceDB
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.strategies.strategy import AbstractStrategy, StrategyFactory
from comb_spec_searcher.strategies.strategy_pack import StrategyPack
from comb_spec_searcher.typing import RuleKey

from .base import RuleDBBase

__all__ = ["RuleDBForgetStrategy"]


class RecomputingDict(MutableMapping[RuleKey, AbstractStrategy]):
    """
    A mapping from rules to strategies that recompute the strategy every time it's
    needed instead of storing it in order to save memory.

    Also in order to save memory we store flat version of the rules, i.e. for a rule
    (a, (b, c,...)) we store (a, b, c,...)
    """

    def __init__(
        self, classdb: ClassDB, strat_pack: StrategyPack, equivdb: EquivalenceDB
    ) -> None:
        self.classdb = classdb
        self.equivdb = equivdb
        self.pack = strat_pack
        self.rules: Set[Tuple[int, ...]] = set()

    @staticmethod
    def _flatten(tuple_: RuleKey) -> Tuple[int, ...]:
        return (tuple_[0],) + tuple_[1]

    @staticmethod
    def _unflatten(tuple_: Tuple[int, ...]) -> RuleKey:
        return (tuple_[0], tuple_[1:])

    def _is_equiv(self, key: RuleKey) -> bool:
        """
        Returns True if the key should be the key of an equivalence rule.
        """
        start = key[0]
        ends = key[1]
        return len(ends) == 1 and self.equivdb.equivalent(start, ends[0])

    def __getitem__(self, key: RuleKey) -> AbstractStrategy:
        if self._flatten(key) not in self.rules:
            raise KeyError(key)
        expect_equiv = self._is_equiv(key)
        possible_labels = (key[0],) + key[1]
        for label, strat in itertools.product(possible_labels, self.pack):
            comb_class = self.classdb.get_class(label)
            if isinstance(strat, StrategyFactory):
                strats_or_rules: Iterable[
                    Union[AbstractRule, AbstractStrategy]
                ] = strat(comb_class)
            else:
                strats_or_rules = [strat]
            for x in strats_or_rules:
                if isinstance(x, AbstractStrategy):
                    try:
                        rule = x(comb_class)
                    except StrategyDoesNotApply:
                        continue
                else:
                    rule = x
                try:
                    start_label = self.classdb.get_label(rule.comb_class)
                    nonempty_children = tuple(
                        c for c in rule.children if not self.classdb.is_empty(c)
                    )
                    end_labels = tuple(
                        sorted(map(self.classdb.get_label, nonempty_children))
                    )
                    if (start_label, end_labels) == key:
                        if expect_equiv and not rule.strategy.can_be_equivalent():
                            continue
                        return rule.strategy
                except StrategyDoesNotApply:
                    pass
        err_message = (
            f"Could not recompute the strategy for the rule {key} with "
            " any of the strategies. Classes are:\n"
        )
        for label in itertools.chain([key[0]], key[1]):
            err_message += f"Label: {label}\n"
            err_message += str(self.classdb.get_class(label)) + "\n"
        raise RuntimeError(err_message)

    def __setitem__(self, key: RuleKey, value: AbstractStrategy) -> None:
        self.rules.add(self._flatten(key))

    def __delitem__(self, key: RuleKey) -> None:
        self.rules.remove(self._flatten(key))

    def __iter__(self) -> Iterator[RuleKey]:
        for rule in self.rules:
            yield self._unflatten(rule)

    def __len__(self) -> int:
        return len(self.rules)

    def __contains__(self, key: object) -> bool:
        return self._flatten(cast(RuleKey, key)) in self.rules


class RuleDBForgetStrategy(RuleDBBase):
    def __init__(self, classdb: ClassDB, strat_pack: StrategyPack) -> None:
        super().__init__()
        self.rules = RecomputingDict(classdb, strat_pack, self.equivdb)

    @property
    def rule_to_strategy(self):
        return self.rules



"""
A database to keep track of equivalent combinatorial classes.

This is done using a union find method. Also, explanations of how combinatorial
classes  are equivalent are maintained.

Based on: https://www.ics.uci.edu/~eppstein/PADS/UnionFind.py

In this file a combinatorial class is represented by a label, which is an
integer, that the classdb gives.
"""

from collections import deque
from typing import Deque, Dict, FrozenSet, List, Set, Union


class EquivalenceDB:
    """
    A database for equivalences. Supports four methods.

    - DB[x] return a name for the set containing x. Each set named by
    arbitrarily chosen member.
    - DB.union(t1, t2, explanation) merges the sets containing t1 and t2 and
    records why t1 and t2 are equivalent.
    - DB.equivalent(t1, t2) returns True, if t1 and t2 are in the same set,
    otherwise returns False.
    - DB.get_relation(t1, t2) returns a string explaining why t1 and t2 are
      equivalent.
    """

    def __init__(self) -> None:
        """Create a new empty equivalent database."""
        self.parents: Dict[int, int] = {}
        self.weights: Dict[int, int] = {}
        self.verified_roots: Set[int] = set()
        self.vertices: Set[FrozenSet[int]] = set()

    def __eq__(self, other: object) -> bool:
        """Check if all information stored is the same."""
        if not isinstance(other, EquivalenceDB):
            return NotImplemented
        return bool(
            self.parents == other.parents
            and self.weights == other.weights
            and self.verified_roots == other.verified_roots
            and self.vertices == other.vertices
        )

    def __getitem__(self, comb_class: int) -> int:
        """Find and return root combinatorial class for the set containing
        comb_class."""
        root = self.parents.get(comb_class)
        if root is None:
            self.parents[comb_class] = comb_class
            self.weights[comb_class] = 1
            return comb_class

        # Find path of combinatorial class leading to root.
        path = [comb_class]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # Compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def union(self, t1: int, t2: int) -> None:
        """Find sets containing t1 and t2 and merge them."""
        if t1 == t2:
            return
        self.vertices.add(frozenset((t1, t2)))
        verified = self.is_verified(t1) or self.is_verified(t2)
        roots = [self[t1], self[t2]]
        heaviest = max([(self.weights[r], r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest
        if verified:
            self.set_verified(t1)

    def equivalent(self, t1: int, t2: int) -> bool:
        """Return True if t1 and t2 are equivalent, False otherwise."""
        return self[t1] == self[t2]

    def set_verified(self, comb_class: int) -> None:
        """Update database that combinatorial classes equivalent to comb_class
        are verified."""
        if not self.is_verified(comb_class):
            self.verified_roots.add(self[comb_class])

    def is_verified(self, comb_class: int) -> bool:
        """Return true if any equivalent combinatorial class is verified."""
        return self[comb_class] in self.verified_roots

    def equivalent_set(self, comb_class: int) -> Set[int]:
        """Return all of the classes equivalent to comb_class."""
        equivalent_classes = set()
        for t in self.parents:
            if self.equivalent(comb_class, t):
                equivalent_classes.add(t)
        return equivalent_classes

    def find_path(self, comb_class: int, other_comb_class: int) -> List[int]:
        """
        BFS for shortest path.

        Used to find shortest explanation of why things are equivalent.
        """
        if not self.equivalent(comb_class, other_comb_class):
            raise KeyError("The classes given are not equivalent.")
        if comb_class == other_comb_class:
            return [comb_class]
        equivalent_comb_classes: Dict[int, int] = {}
        reverse_map: Dict[int, int] = {}

        for x in self.parents:
            n = len(equivalent_comb_classes)
            if self.equivalent(comb_class, x):
                equivalent_comb_classes[x] = n
                reverse_map[n] = x

        adjacency_list: List[List[int]] = [
            [] for i in range(len(equivalent_comb_classes))
        ]
        for (t1, t2) in self.vertices:
            if self.equivalent(t1, comb_class):
                u = equivalent_comb_classes[t1]
                v = equivalent_comb_classes[t2]
                adjacency_list[u].append(v)
                adjacency_list[v].append(u)

        dequeue: Deque[int] = deque()

        start = equivalent_comb_classes[comb_class]
        end = equivalent_comb_classes[other_comb_class]

        n = len(equivalent_comb_classes)

        dequeue.append(start)
        visited = [False for i in range(n)]
        neighbour: List[Union[int, None]] = [None for i in range(n)]
        while dequeue:
            u = dequeue.popleft()
            if u == end:
                break
            if visited[u]:
                continue
            visited[u] = True

            for v in adjacency_list[u]:
                if not visited[v]:
                    dequeue.append(v)
                    neighbour[v] = u

        path = [reverse_map[end]]
        while neighbour[end] is not None:
            assert isinstance(end, int), "something went wrong"
            a = neighbour[end]
            assert isinstance(a, int), "something went wrong"
            t = reverse_map[a]
            path.append(t)
            end = a

        return path[::-1]



from .comb_spec_searcher import CombinatorialSpecificationSearcher
from .combinatorial_class import CombinatorialClass, CombinatorialObject
from .specification import CombinatorialSpecification
from .specification_drawer import ForestSpecificationDrawer, SpecificationDrawer
from .strategies import (
    AtomStrategy,
    CartesianProduct,
    CartesianProductStrategy,
    Constructor,
    DisjointUnion,
    DisjointUnionStrategy,
    Strategy,
    StrategyFactory,
    StrategyPack,
    SymmetryStrategy,
    VerificationStrategy,
)

__all__ = [
    "CombinatorialSpecificationSearcher",
    "CombinatorialClass",
    "CombinatorialObject",
    "CombinatorialSpecification",
    "ForestSpecificationDrawer",
    "SpecificationDrawer",
    "AtomStrategy",
    "CartesianProduct",
    "CartesianProductStrategy",
    "Constructor",
    "DisjointUnion",
    "DisjointUnionStrategy",
    "Strategy",
    "StrategyFactory",
    "StrategyPack",
    "SymmetryStrategy",
    "VerificationStrategy",
]



"""
An abstract class for a CombinatorialClass.
"""
import abc
from functools import reduce
from importlib import import_module
from operator import mul
from typing import Counter, DefaultDict, Dict, Generic, Iterator, List, Tuple, Type

from sympy import Expr, Number, var

from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Objects,
    Parameters,
    Terms,
)

__all__ = ("CombinatorialClass", "CombinatorialObject")


class CombinatorialObject(abc.ABC):
    """Abstract class for a combinatorial object."""

    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    def size(self) -> int:
        """Return the size of the object"""
        return len(self)


class CombinatorialClass(Generic[CombinatorialObjectType], abc.ABC):
    """
    Base class for CombSpecSearcher combinatorial class

    This is a base combinatorial class that CombinatorialSpecificationSearcher
    works with. Combinatorial classes to be used with the searcher should
    inherit from this class.
    """

    @abc.abstractmethod
    def is_empty(self) -> bool:
        """Return True if there are no object of any lengths."""
        return False

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, d: dict) -> "CombinatorialClass":
        """Return the combinatorial class from the json dict representation."""
        module = import_module(d["class_module"])
        StratClass: Type["CombinatorialClass"] = getattr(module, d["comb_class"])
        assert issubclass(
            StratClass, CombinatorialClass
        ), "Not a valid CombinatorialClass"
        return StratClass.from_dict(d)  # type: ignore

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the combinatorial class."""
        c = self.__class__
        return {
            "class_module": c.__module__,
            "comb_class": c.__name__,
        }

    def get_genf(self, *args, **kwargs) -> Expr:
        """Return the generating function for the combinatorial class"""
        raise NotImplementedError(
            (
                "If you want to use the 'get_genf' function"
                "for a proof tree then you must implement"
                " 'get_genf' for verified combinatorial "
                "classes."
            )
        )

    @property
    def extra_parameters(self) -> Tuple[str, ...]:
        """
        Return the parameters used to get the enumeration of the
        class. It is assumed that the order is consistent for a combinatorial class.
        """
        return tuple()

    def get_minimum_value(self, parameter: str) -> int:
        """
        Return the minimum value that can be taken by the parameter. This is
        required to be implemented if to use CartesianProduct on multiple
        variables.
        """
        raise NotImplementedError(
            "You need to implement the minimum value a parameter can take on "
            "your CombinatorialClass to use the CartesianProduct constructor."
        )

    def possible_parameters(self, n: int) -> Iterator[Dict[str, int]]:
        """
        Return all the possible values the extra parameters could take for
        the given value of n.
        """
        if self.extra_parameters:
            raise NotImplementedError(
                "You need to implement the possible parameters on your "
                "CombinatorialClass in order to use various methods, including "
                "sanity checking."
            )
        yield dict()

    def get_parameters(self, obj: CombinatorialObjectType) -> Parameters:
        """
        Returns the appropriate parameter tuple of the given object.
        """
        if self.extra_parameters:
            raise NotImplementedError(
                "You need to implement this method for combclass with extra parameters"
            )
        return tuple()

    def objects_of_size(
        self, n: int, **parameters: int
    ) -> Iterator[CombinatorialObjectType]:
        """
        Returns an iterable of combinatorial objects of a given size.

        If the combinatorial class has extra parameters and None are given, the iterator
        should go over all objects of size n.
        """
        raise NotImplementedError

    def get_terms(self, n: int) -> Terms:
        terms: Terms = Counter()
        for obj in self.objects_of_size(n):
            param = self.get_parameters(obj)
            terms[param] += 1
        return terms

    def get_objects(self, n: int) -> Objects:
        objects: Objects = DefaultDict(list)
        for obj in self.objects_of_size(n):
            param = self.get_parameters(obj)
            objects[param].append(obj)  # pylint: disable=invalid-sequence-index
        return objects

    def initial_conditions(self, check: int = 6) -> List[Expr]:
        """
        Returns a list with the initial conditions to size `check` of the
        CombinatorialClass.
        """

        def monomial(parameters: Dict[str, int]) -> Expr:
            return reduce(
                mul, [var(k) ** val for k, val in parameters.items()], Number(1)
            )

        return [
            sum(
                sum(Number(1) for _ in self.objects_of_size(n, **parameters))
                * monomial(parameters)
                for parameters in self.possible_parameters(n)
            )
            for n in range(check + 1)
        ]

    def is_atom(self):
        """Returns True if the combinatorial class contains a single object."""
        raise NotImplementedError(
            "To use the CartesianProduct constructor, 'is_atom' and "
            "'minimum_size_of_object' must be implemented."
        )

    def minimum_size_of_object(self) -> int:
        """Return the size of the smallest object in the combinatorial class.
        Note, for productivity reasons, you must at least return 1, if this
        should be greater than 0."""
        raise NotImplementedError(
            "To use the CartesianProduct constructor, 'is_atom' and "
            "'minimum_size_of_object' must be implemented."
        )

    def to_bytes(self) -> bytes:
        """Return a compressed version of the class in the form of a 'bytes'
        object. If you are having memory issues then implement this function
        and the from_bytes function such that
        'cls.from_bytes(self.to_bytes()) == self'"""
        raise NotImplementedError

    def to_html_representation(self) -> str:
        """Returns a string containing html representation for this class"""
        raise NotImplementedError("Use string method instead")

    @classmethod
    def from_bytes(
        cls: Type[CombinatorialClassType], b: bytes
    ) -> CombinatorialClassType:
        """
        Return decompressed class from the bytes object returned by the
        'to_bytes' function. If you are having memory issues then implement
        this function and the to_bytes function such that
        'cls.from_bytes(self.to_bytes()) == self'
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __init__(self):
        return

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        return False

    @abc.abstractmethod
    def __hash__(self) -> int:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    @abc.abstractmethod
    def __str__(self) -> str:
        pass

    def _get_class_name(self) -> str:
        """Return the name of the class."""
        return type(self).__name__



"""
A proof tree class.

This can be used to get the generating function for the class.

The class is only used for reverse compatibility with ComboPal. You should use
the Specification class.
"""
import json
import warnings
from typing import Iterable, List, Optional

from .combinatorial_class import CombinatorialClass
from .specification import CombinatorialSpecification
from .strategies.constructor import CartesianProduct, Constructor, DisjointUnion
from .strategies.rule import EquivalencePathRule, Rule, VerificationRule

__all__ = ("ProofTree",)


class ProofTreeNode:
    """Class is deprecated - use CombinatorialSpecification."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        label: int,
        eqv_path_labels: Iterable[int],
        eqv_path_comb_classes: Iterable[CombinatorialClass],
        eqv_explanations: Iterable[str] = None,
        children: Iterable["ProofTreeNode"] = None,
        strategy_verified: bool = False,
        decomposition: bool = False,
        disjoint_union: bool = False,
        recursion: bool = False,
        formal_step: str = "",
    ):
        self.label = label
        self.eqv_path_labels = list(eqv_path_labels)
        self.eqv_path_comb_classes = list(eqv_path_comb_classes)
        if eqv_explanations is not None:
            self.eqv_explanations = list(eqv_explanations)
        else:
            self.eqv_explanations = []
        self.children = list(children) if children is not None else []
        self.strategy_verified = strategy_verified
        self.decomposition = decomposition
        self.disjoint_union = disjoint_union
        self.recursion = recursion
        self.formal_step = formal_step

    def to_jsonable(self):
        """Class is deprecated - use CombinatorialSpecification."""
        output = dict()
        output["label"] = self.label
        output["eqv_path_labels"] = list(self.eqv_path_labels)
        output["eqv_path_comb_classes"] = [
            x.to_jsonable() for x in self.eqv_path_comb_classes
        ]
        output["eqv_explanations"] = list(self.eqv_explanations)
        output["children"] = [child.to_jsonable() for child in self.children]
        output["strategy_verified"] = self.strategy_verified
        output["decomposition"] = self.decomposition
        output["disjoint_union"] = self.disjoint_union
        output["recursion"] = self.recursion
        output["formal_step"] = self.formal_step
        return output

    @classmethod
    def from_dict(cls, combclass, jsondict):
        """Class is deprecated - use CombinatorialSpecification."""
        if "eqv_path_objects" in jsondict:
            warnings.warn(
                (
                    "The 'eqv_path_objects' label is deprecated. You "
                    "should change this to 'eqv_path_comb_classes"
                    " in the future."
                ),
                DeprecationWarning,
                stacklevel=2,
            )
            jsondict["eqv_path_comb_classes"] = jsondict["eqv_path_objects"]
        return cls(
            label=jsondict["label"],
            eqv_path_labels=jsondict["eqv_path_labels"],
            eqv_path_comb_classes=[
                combclass.from_dict(x) for x in jsondict["eqv_path_comb_classes"]
            ],
            eqv_explanations=jsondict["eqv_explanations"],
            children=[
                ProofTreeNode.from_dict(combclass, child)
                for child in jsondict["children"]
            ],
            strategy_verified=jsondict["strategy_verified"],
            decomposition=jsondict["decomposition"],
            disjoint_union=jsondict["disjoint_union"],
            recursion=jsondict["recursion"],
            formal_step=jsondict["formal_step"],
        )

    @classmethod
    def from_json(cls, combclass, jsonstr):
        """Class is deprecated - use CombinatorialSpecification."""
        warnings.warn("ProofTreeNode is deprecated, use CombinatorialSpecification")
        jsondict = json.loads(jsonstr)
        return cls.from_dict(combclass, jsondict)


class ProofTree:
    """Class is deprecated - use CombinatorialSpecification."""

    def __init__(self, root):
        warnings.warn(
            (
                "The ProofTree class is deprecated."
                " Use CombinatorialSpecification instead."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        if not isinstance(root, ProofTreeNode):
            raise TypeError("Root must be a ProofTreeNode.")
        self.root = root

    def to_jsonable(self):
        """Class is deprecated - use CombinatorialSpecification."""
        return {"root": self.root.to_jsonable()}

    @classmethod
    def from_dict(cls, combclass, jsondict):
        """Class is deprecated - use CombinatorialSpecification."""
        root = ProofTreeNode.from_dict(combclass, jsondict["root"])
        return cls(root)

    @classmethod
    def from_json(cls, combclass, jsonstr):
        """Class is deprecated - use CombinatorialSpecification."""
        jsondict = json.loads(jsonstr)
        return cls.from_dict(combclass, jsondict)

    @classmethod
    def from_specification(cls, spec: CombinatorialSpecification) -> "ProofTree":
        """Return a ProofTree from a CombinatorialSpecification."""
        seen: List[Rule] = list()

        def proof_tree_node(comb_class: CombinatorialClass) -> ProofTreeNode:
            rule = spec.get_rule(comb_class)
            # Setting up the equivalence path
            if isinstance(rule, EquivalencePathRule):
                eqv_path_labels: List[int] = []
                eqv_path_comb_classes: List[CombinatorialClass] = []
                eqv_explanations: List[str] = []
                for path_class, eqv_rule in rule.eqv_path_rules():
                    eqv_path_labels.append(spec.get_label(path_class))
                    eqv_path_comb_classes.append(path_class)
                    eqv_explanations.append(eqv_rule.formal_step)
                eqv_path_comb_classes.append(rule.children[0])
                rule = spec.rules_dict[rule.children[0]]
            else:
                eqv_path_labels = [spec.get_label(rule.comb_class)]
                eqv_path_comb_classes = [rule.comb_class]
                eqv_explanations = []
            if isinstance(rule, VerificationRule):
                return ProofTreeNode(
                    label=spec.get_label(rule.comb_class),
                    eqv_path_labels=eqv_path_labels,
                    eqv_path_comb_classes=eqv_path_comb_classes,
                    eqv_explanations=eqv_explanations,
                    children=[],
                    strategy_verified=True,
                    formal_step=rule.formal_step,
                )
            if isinstance(rule, Rule):
                if rule in seen:
                    # Setting up recursion node
                    return ProofTreeNode(
                        label=spec.get_label(rule.comb_class),
                        eqv_path_labels=eqv_path_labels,
                        eqv_path_comb_classes=eqv_path_comb_classes,
                        eqv_explanations=eqv_explanations,
                        children=[],
                        strategy_verified=False,
                        recursion=True,
                        formal_step="recurse",
                    )
                seen.append(rule)
                children = [proof_tree_node(c) for c in rule.children]
                try:
                    constructor: Optional[Constructor] = rule.constructor
                except NotImplementedError:
                    constructor = None
                return ProofTreeNode(
                    label=spec.get_label(rule.comb_class),
                    eqv_path_labels=eqv_path_labels,
                    eqv_path_comb_classes=eqv_path_comb_classes,
                    eqv_explanations=eqv_explanations,
                    children=children,
                    decomposition=isinstance(constructor, CartesianProduct),
                    disjoint_union=isinstance(constructor, DisjointUnion),
                    strategy_verified=False,
                    formal_step=rule.formal_step,
                )
            raise ValueError(f"Don't know what to do with the rule class of {rule}")

        return ProofTree(proof_tree_node(spec.root))



"""
Finds and returns a combinatorial specification, that we call a proof tree.
"""
import time
from collections import defaultdict, deque
from copy import deepcopy
from itertools import chain, product
from random import choice, shuffle
from typing import Dict, FrozenSet, Iterator, List, Optional, Sequence, Set, Tuple

from comb_spec_searcher.typing import RulesDict

__all__ = ("prune", "proof_tree_generator_dfs", "proof_tree_generator_bfs")


class Node:
    """A node for a proof tree."""

    def __init__(self, n: int, children: Optional[List["Node"]] = None):
        if children is None:
            children = []
        self.label = n
        self.children = children

    def labels(self) -> Set[int]:
        """Return the set of all labels in the proof tree."""
        res = set([self.label])
        res.update(chain.from_iterable(node.labels() for node in self.children))
        return res

    def nodes(self) -> Iterator["Node"]:
        """Yield all nodes in the proof tree."""
        yield self
        for node in self.children:
            yield from node.nodes()

    def __str__(self) -> str:
        return "".join(["(", str(self.label), *map(str, self.children), ")"])

    def __len__(self) -> int:
        """Return the number nodes in the proof tree."""
        return 1 + sum(len(c) for c in self.children)


def prune(rdict: RulesDict) -> None:
    """
    Prune all nodes not in a combinatorial specification. This changes rdict
    in place.
    """
    changed = True
    while changed:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if any(x not in rdict for x in rule):
                    rule_set.remove(rule)
                    changed = True
                if not rule_set:
                    del rdict[k]


def iterative_prune(rules_dict: RulesDict, root: Optional[int] = None) -> RulesDict:
    """Prune all nodes not iteratively verifiable."""
    verified_labels: Set[int] = set()
    if root is not None:
        verified_labels.add(root)
    rdict = deepcopy(rules_dict)
    new_rules_dict: RulesDict = defaultdict(set)
    while True:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if all(x in verified_labels for x in rule):
                    changed = True
                    verified_labels.add(k)
                    new_rules_dict[k].add(rule)
                    rdict[k].remove(rule)
            if not rule_set:
                del rdict[k]
        if not changed:
            break
    return new_rules_dict


def proof_tree_dfs(rules_dict: RulesDict, root: int, seen: Optional[Set[int]] = None):
    """Return random proof tree found by depth first search."""
    if seen is None:
        seen = set()
    seen = seen.copy()
    if root in rules_dict:
        rule_set = rules_dict[root]
        root_node = Node(root)
        if root in seen or () in rule_set:
            seen.add(root)
            return seen, root_node
        seen.add(root)
        rule = choice(list(rule_set))
        visited, trees = all_proof_trees_dfs(rules_dict, rule, seen)
        root_node.children = trees
        return visited, root_node


def all_proof_trees_dfs(
    rules_dict: RulesDict, roots: Sequence[int], seen: Optional[Set[int]] = None
) -> Tuple[Set[int], List[Node]]:
    """Return all labels which have been seen, together with all of the trees
    using the given roots.."""
    if seen is None:
        seen = set()
    if not roots:
        return seen, []
    root, roots = roots[0], roots[1:]
    seen1, tree = proof_tree_dfs(rules_dict, root, seen)
    seen2, trees = all_proof_trees_dfs(rules_dict, roots, seen1)
    return seen1.union(seen2), [tree] + trees


def iterative_proof_tree_bfs(rules_dict: RulesDict, root: int) -> Node:
    """Takes in a iterative pruned rules_dict and returns iterative proof
    tree."""
    root_node = Node(root)
    queue = deque([root_node])
    while queue:
        v = queue.popleft()
        rule = sorted(rules_dict[v.label])[0]
        if not rule == ():
            children = [Node(i) for i in rule]
            queue.extend([child for child in children if not child.label == root])
            v.children = children
    return root_node


def random_proof_tree(rules_dict: RulesDict, root: int) -> Node:
    """Return random tree found by breadth first search."""
    seen: Set[int] = set()
    root_node = Node(root)
    queue = deque([root_node])
    while queue:
        v = queue.popleft()
        rule = choice(list(rules_dict[v.label]))
        if not (v.label in seen or rule == ()):
            children = [Node(i) for i in rule]
            shuffle(children)
            queue.extend(children)
            v.children = children
        seen.add(v.label)
    return root_node


def smallish_random_proof_tree(
    rules_dict: RulesDict, root: int, minimization_time_limit: float
) -> Node:
    """
    Searches a rule_dict known to contain at least one specification for a
    small specification. Spends minimization_time_limit seconds searching.
    """
    start_time = time.time()
    smallest_so_far = random_proof_tree(rules_dict, root=root)
    smallest_size = len(smallest_so_far)
    while time.time() - start_time < minimization_time_limit:
        next_tree = random_proof_tree(rules_dict, root=root)
        next_tree_size = len(next_tree)
        if next_tree_size < smallest_size:
            smallest_so_far = next_tree
            smallest_size = next_tree_size
    return smallest_so_far


def proof_tree_generator_bfs(rules_dict: RulesDict, root: int) -> Iterator[Node]:
    """A generator for all proof trees using breadth first search.
    N.B. The rules_dict is assumed to be pruned.
    """

    def _bfs_helper(root_label: int, seen: FrozenSet[int]):
        if root_label in seen:
            yield Node(root_label)
            return
        next_seen = seen.union((root_label,))
        for rule in rules_dict[root_label]:
            for children in product(
                *[_bfs_helper(child_label, next_seen) for child_label in rule]
            ):
                root_node = Node(root_label)
                root_node.children = children
                yield root_node

    sorted_rules_dict = {
        start: tuple(sorted(ends)) for start, ends in rules_dict.items()
    }

    if root in sorted_rules_dict:
        yield from _bfs_helper(root, frozenset())


def proof_tree_generator_dfs(
    rules_dict: RulesDict, root: int, maximum: Optional[int] = None
) -> Iterator[Node]:
    """A generator for all proof trees using depth first search.
    N.B. The rules_dict is assumed to be pruned.
    """

    def _dfs_tree(
        root_label: int, seen: FrozenSet[int], maximum: int = None
    ) -> Iterator[Tuple[FrozenSet[int], Node]]:
        if maximum is not None and maximum <= 0:
            return
        if root_label in seen:
            yield seen, Node(root_label)
            return
        seen = seen.union((root_label,))
        for rule in sorted_rules_dict[root_label]:
            if rule == ():
                yield seen, Node(root_label)
            else:
                for new_seen, children in _dfs_forest(rule, seen, maximum):
                    root_node = Node(root_label)
                    root_node.children = children
                    yield new_seen, root_node

    def _dfs_forest(
        root_labels: Sequence[int], seen: FrozenSet[int], maximum: Optional[int] = None
    ) -> Iterator[Tuple[FrozenSet[int], List[Node]]]:
        if maximum is not None and maximum <= 0:
            return
        if not root_labels:
            yield seen, []
        else:
            root, roots = root_labels[0], root_labels[1:]
            new_max = maximum - len(root_labels) + 1 if maximum is not None else None
            for seen1, tree in _dfs_tree(root, seen, new_max):
                length = len(tree)
                new_maximum = maximum - length if maximum is not None else None
                for seen2, trees in _dfs_forest(roots, seen1, new_maximum):
                    actual_length = length + sum(len(t) for t in trees)
                    if maximum is not None and actual_length < maximum:
                        yield seen1.union(seen2), [tree] + trees

    sorted_rules_dict = {
        start: tuple(sorted(ends)) for start, ends in rules_dict.items()
    }

    if root in sorted_rules_dict:
        for _, tree in _dfs_tree(root, frozenset(), maximum):
            yield tree


def iterative_proof_tree_finder(rules_dict: RulesDict, root: int) -> Node:
    """Finds an iterative proof tree for root, if one exists."""
    trees: Dict[int, Node] = {}

    def get_tree(start):
        if start == root:
            return Node(start)
        if start in trees:
            return trees[start]
        raise KeyError("{} is not in trees".format(start))

    def create_tree(start, end):
        if start in trees:
            return
        root = Node(start)
        children = [get_tree(i) for i in end]
        root.children = children
        trees[start] = root

    verified_labels = set()
    if root is not None:
        verified_labels.add(root)
    rdict = deepcopy(rules_dict)
    new_rules_dict: RulesDict = defaultdict(set)
    while True:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if all(x in verified_labels for x in rule):
                    changed = True
                    verified_labels.add(k)
                    new_rules_dict[k].add(rule)
                    create_tree(k, rule)
                    rdict[k].remove(rule)
            if not rule_set:
                del rdict[k]
        if not changed:
            break
    if root in trees:
        return trees[root]
    raise ValueError("{} has no tree in rules_dict".format(root))



"""
A database for combinatorial class found. It gives each combinatorial class
a unique label. Each combinatorial class object is compressed, and decompressed
using the CombinatorialClass 'to_bytes' and 'from_bytes' method if these are
implemented.

Contains information about if combinatorial classes have been found by
if is_empty has been checked.
"""

import zlib
from typing import Dict, Generic, Iterator, Optional, Type, cast

from comb_spec_searcher.typing import ClassKey, CombinatorialClassType, Key


class Info:
    """
    Information about a combinatorial class.
        - the class,
        - the label,
        - is it empty?
    """

    def __init__(self, comb_class: ClassKey, label: int, empty: Optional[bool] = None):
        self.comb_class = comb_class
        self.label = label
        self.empty = empty

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Info):
            return NotImplemented
        return (
            self.comb_class == self.comb_class
            and self.label == self.label
            and self.empty == self.empty
        )


class ClassDB(Generic[CombinatorialClassType]):
    """
    A database for combinatorial classes.

    Each class is given a unique integer label. The key to the database is
    both the unique integer label and the ckass. It supports the following
    methods.
    - DB.add(class) will label the class and add it to the database.
    - DB.get_class(key) return the class with the given key.
    - DB.get_label(key) return the label of the given key.
    - DB.set_empty(key) will set empty to be true for key.
    - DB.is_empty(key) will return True if the key is empty, False
    otherwise.
    """

    def __init__(self, combinatorial_class: Type[CombinatorialClassType]):
        self.class_to_info: Dict[ClassKey, Info] = {}
        self.label_to_info: Dict[int, Info] = {}
        self.combinatorial_class = combinatorial_class

    def __iter__(self) -> Iterator[int]:
        """
        Iterator of labels.
        """
        for key in self.label_to_info:
            yield key

    def __contains__(self, key: Key) -> bool:
        """
        Return true if the the key is already in the database.
        """
        if isinstance(key, self.combinatorial_class):
            comb_class = self._compress(key)
            info = self.class_to_info.get(comb_class)
        elif isinstance(key, int):
            info = self.label_to_info.get(key)
        else:
            raise ValueError("Invalid key")
        return info is not None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassDB):
            return NotImplemented
        return bool(
            self.class_to_info == other.class_to_info
            and self.label_to_info == other.label_to_info
        )

    def add(self, comb_class: ClassKey, compressed: bool = False) -> None:
        """
        Add a combinatorial class to the database.
        """
        if not compressed and not isinstance(comb_class, self.combinatorial_class):
            raise TypeError(
                ("Trying to add something that isn't a" "CombinatorialClass.")
            )
        if not compressed:
            assert isinstance(
                comb_class, self.combinatorial_class
            ), "trying to add non combinatorial class to database"
            compressed_class = self._compress(comb_class)
        else:
            compressed_class = comb_class
        if compressed_class not in self.class_to_info:
            label = len(self.class_to_info)
            info = Info(compressed_class, label)
            self.class_to_info[compressed_class] = info
            self.label_to_info[label] = info
        else:
            label = self.class_to_info[compressed_class].label

    def _get_info(self, key: Key) -> Info:
        """
        Return Info for given key.
        """
        if isinstance(key, self.combinatorial_class):
            compressed_key = self._compress(key)
            info = self.class_to_info.get(compressed_key)
            if info is None:
                self.add(compressed_key, compressed=True)
                info = self.class_to_info[compressed_key]
        elif isinstance(key, int):
            info = self.label_to_info.get(key)
            if info is None:
                raise KeyError("Key not in ClassDB.")
        else:
            raise TypeError(
                "ClassDB only accepts"
                "CombinatorialClass and will decompress with"
                "{}.".format(self.combinatorial_class)
            )
        return info

    def _compress(self, key: CombinatorialClassType) -> ClassKey:
        """
        Return compressed version of combinatorial class.
        """
        # pylint: disable=no-self-use
        try:
            return zlib.compress(key.to_bytes(), 9)
        except NotImplementedError:
            # to use compression you should implement a 'to_bytes' function.
            return key

    def _decompress(self, key: ClassKey) -> CombinatorialClassType:
        """
        Return decompressed version of compressed combinatorial class.
        """
        try:
            assert isinstance(key, bytes)
            return cast(
                CombinatorialClassType,
                self.combinatorial_class.from_bytes(zlib.decompress(key)),
            )
        except (AssertionError, NotImplementedError):
            # to use compression you should implement a 'from_bytes' function.
            assert isinstance(
                key, self.combinatorial_class
            ), "you must implement a 'from_bytes' function to use compression"
            return key

    def get_class(self, key: Key) -> CombinatorialClassType:
        """
        Return combinatorial class of key.
        """
        info = self._get_info(key)
        return self._decompress(info.comb_class)

    def get_label(self, key: Key) -> int:
        """
        Return label of key.
        """
        info = self._get_info(key)
        return info.label

    def is_empty(
        self, comb_class: CombinatorialClassType, label: Optional[int] = None
    ) -> bool:
        """
        Return True if combinatorial class is empty set, False if not.
        Return None if status not set.
        """
        if label is None:
            info = self._get_info(comb_class)
            label = info.label
        else:
            info = self._get_info(label)
        empty = info.empty
        if empty is None:
            if not isinstance(comb_class, self.combinatorial_class):
                comb_class = self.get_class(comb_class)
            empty = self._is_empty(comb_class)
            self.set_empty(label, empty)
        return bool(empty)

    def _is_empty(self, comb_class: CombinatorialClassType) -> bool:
        if not isinstance(comb_class, self.combinatorial_class):
            comb_class = self.get_class(comb_class)
        return comb_class.is_empty()

    def set_empty(self, key: Key, empty: bool = True) -> None:
        """
        Update database about comb class being empty.
        """
        info = self._get_info(key)
        info.empty = empty

    def status(self) -> str:
        """
        Return a string with the current status of the run.
        """
        status = "ClassDB status:\n"
        status += "\tTotal number of combinatorial classes found is"
        status += f" {len(self.label_to_info):,d}"
        return status



"""Some useful miscellaneous functions used througout the package."""
import functools
import operator
import os
import re
import sys
import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

import psutil
import sympy

from comb_spec_searcher.exception import TaylorExpansionError

if TYPE_CHECKING:
    from comb_spec_searcher import CombinatorialSpecificationSearcher


Func = TypeVar("Func", bound=Callable[..., Any])


class cssmethodtimer:
    """This is a decorator for counting and timing function calls."""

    def __init__(self, explanation: str):
        self.explanation = explanation

    def __call__(self, func: Func) -> Func:
        def inner(css: "CombinatorialSpecificationSearcher", *args, **kwargs):
            start = time.time()
            res = func(css, *args, **kwargs)
            css.func_times[self.explanation] += time.time() - start
            css.func_calls[self.explanation] += 1
            return res

        return cast(Func, inner)


class cssiteratortimer:
    """This is a decorator for counting and timing function calls."""

    def __init__(self, explanation: str):
        self.explanation = explanation

    def __call__(self, func: Func) -> Func:
        def inner(css: "CombinatorialSpecificationSearcher", *args, **kwargs):
            key = self.explanation
            if self.explanation == "_expand_class_with_strategy":
                key = str(args[1])
            css.func_calls[key] += 1
            start = time.time()
            for res in func(css, *args, **kwargs):
                css.func_times[key] += time.time() - start
                yield res
                start = time.time()
            css.func_times[key] += time.time() - start

        return cast(Func, inner)


class RecursionLimit:
    """
    A context manager to momentarily increase the recursion limit.
    """

    def __init__(self, limit: int):
        self.curr_limit = sys.getrecursionlimit()
        if self.curr_limit < limit:
            self.limit = limit
        else:
            self.limit = self.curr_limit

    def __enter__(self) -> None:
        sys.setrecursionlimit(self.limit)

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        sys.setrecursionlimit(self.curr_limit)


def check_poly(min_poly, initial, root_initial=None, root_func=None):
    """Return True if this is a minimum polynomial for the generating
    function F with the given initial terms. Input is a polynomial in F,
    and initial terms."""
    F = sympy.Symbol("F")
    x = sympy.var("x")
    init_poly = 0
    for i, coeff in enumerate(initial):
        init_poly += coeff * x ** i
    if root_initial is not None:
        root_poly = 0
        for i, coeff in enumerate(root_initial):
            root_poly += coeff * x ** i
    verification = min_poly.subs({F: init_poly})
    if root_initial is not None:
        verification = verification.subs({root_func: root_poly})
    verification = verification.expand()
    verification = verification.series(x, n=len(initial)).removeO()
    verification = (verification + sympy.O(x ** (len(initial) - 1))).removeO()
    return verification == 0


def check_equation(equation, initial, root_initial=None, root_func=None):
    """Return True if an equation in terms of the generating function F and x
    is satisfied."""
    F = sympy.Symbol("F")
    solutions = sympy.solve(
        equation, F, dict=True, cubics=False, quartics=False, quintics=False
    )
    for solution in solutions:
        genf = solution[F]
        try:
            expansion = taylor_expand(genf, len(initial) - 1)
        except TaylorExpansionError:
            continue
        if initial == expansion:
            return True
    return False


def get_solution(equation, initial):
    """Return solution of equation in F and x with the given initial
    conditions."""
    F = sympy.Symbol("F")
    solutions = sympy.solve(
        equation, F, dict=True, cubics=False, quartics=False, quintics=False
    )
    for solution in solutions:
        genf = solution[F]
        try:
            expansion = taylor_expand(genf, len(initial) - 1)
        except TaylorExpansionError:
            continue
        if initial == expansion:
            return genf


def taylor_expand(genf, n: int = 10):
    """Taylor expand the given expression in x."""
    x = sympy.var("x")
    try:
        num, den = genf.as_numer_denom()
        num = num.expand()
        den = den.expand()
        genf = num / den
        ser = sympy.Poly(genf.series(n=n + 1, x=x).removeO(), x)
        res = ser.all_coeffs()
        res = res[::-1] + [0] * (n + 1 - len(res))
    except Exception as e:
        raise TaylorExpansionError from e
    return res


def pretty_print_equations(root_func, count, eqs) -> str:
    s = "The system of {} equations\n".format(len(eqs))
    s += "root_func := {}:\n".format(str(root_func))
    s += "eqs := [\n"
    s += ",\n".join("{} = {}".format(str(eq.lhs), str(eq.rhs)) for eq in eqs)
    s += "\n]:\n"
    s += "count := {}:".format(list(count))
    if all(len(eq.lhs.args) == 1 for eq in eqs):
        s = s.replace("(x)", "")
    return s


def maple_equations(root_func, count, eqs) -> str:
    """
    Convert a systems of equations to version that can be copy pasted to maple.
    """
    s = f"root_func := {sympy_expr_to_maple(root_func)}:\n"
    s += "eqs := [\n"
    s += ",\n".join(map(sympy_expr_to_maple, eqs))
    s += "\n]:\n"
    s += "count := {}:".format(list(count))
    return s


def sympy_expr_to_maple(expr):
    """
    Convert a sympy expression to a maple string.
    """
    if isinstance(expr, sympy.Eq):
        return f"{sympy_expr_to_maple(expr.lhs)} = {sympy_expr_to_maple(expr.rhs)}"
    if isinstance(expr, sympy.core.add.Add):
        head, tail = expr.as_two_terms()
        return f"({sympy_expr_to_maple(head)} + {sympy_expr_to_maple(tail)})"
    if isinstance(expr, sympy.core.add.Mul):
        head, tail = expr.as_two_terms()
        return f"({sympy_expr_to_maple(head)} * {sympy_expr_to_maple(tail)})"
    if isinstance(expr, sympy.core.power.Pow):
        base, exp = expr.as_base_exp()
        return f"({sympy_expr_to_maple(base)}**{sympy_expr_to_maple(exp)})"
    if isinstance(expr.__class__, sympy.core.function.UndefinedFunction):
        if "NOTIMPLEMENTED" in str(expr):
            return "NOTIMPLEMENTED"
        split = re.compile(r"F_([0-9]+)\((.*)\)")
        assert split.match(repr(expr)) is not None, expr
        label = split.match(repr(expr)).group(1)
        args = map(sympy.sympify, split.match(repr(expr)).group(2).split(", "))
        content = f"{label}, " + ", ".join(map(sympy_expr_to_maple, args))
        return f"F[{content}]"
    if isinstance(expr, sympy.core.symbol.Symbol):
        symb = str(expr)
        if "_" in symb:
            var, label = symb.split("_")
            return f"{var}[{label}]"
        return symb
    if expr.is_number:
        return f"({expr})"
    raise NotImplementedError(str(expr))


def compositions(
    n: int, k: int, min_sizes: Tuple[int, ...], max_sizes: Tuple[Optional[int], ...]
) -> Iterator[Tuple[int, ...]]:
    """
    Iterator over all composition of n in k parts with the given max_sizes and
    min_sizes.
    """
    if (
        n < 0
        or k <= 0
        or n < sum(min_sizes)
        or (
            all(s is not None for s in max_sizes)
            and sum(cast(Tuple[int, ...], max_sizes)) < n
        )
    ):
        return
    if k == 1:
        assert (max_sizes[0] is None or max_sizes[0] >= n) and n >= min_sizes[0]
        yield (n,)
        return
    max_size = max_sizes[0] if max_sizes[0] is not None else n
    for i in range(min_sizes[0], max_size + 1):
        yield from map(
            (i,).__add__, compositions(n - i, k - 1, min_sizes[1:], max_sizes[1:])
        )


def prod(values: Iterable[int]) -> int:
    """Compute the product of the integers."""
    return functools.reduce(operator.mul, values)


def nice_pypy_mem(mem: str) -> str:
    """Improves formatting of the memory statistics produced by pypy's
    garbage collector (which are provided as strings)"""
    return mem.replace("KB", " KiB").replace("MB", " MiB").replace("GB", " GiB")


def get_mem() -> int:
    """Return memory used by CombSpecSearcher - note this is actually the
    memory usage of the process that the instance of CombSpecSearcher was
    invoked."""
    return int(psutil.Process(os.getpid()).memory_info().rss)


def size_to_readable(size: int) -> str:
    """Convert a size in bytes to a human readable value in KiB, MiB, or
    GiB"""
    if size / 1024 ** 2 < 1:
        return str(round(size / 1024)) + " KiB"
    if size / 1024 ** 3 < 1:
        return str(round(size / 1024 ** 2, 1)) + " MiB"
    return str(round(size / 1024 ** 3, 3)) + " GiB"



"""
A combinatorial specification is a set rules of the form a -> b1, ..., bk
where each of the bi appear exactly once on the left hand side of some rule.
"""
from copy import copy
from functools import reduce
from operator import mul
from typing import Dict, Generic, Iterable, Iterator, List, Sequence, Set, Tuple, Union

import sympy
from logzero import logger
from sympy import Eq, Expr, Function, Number, solve, var

from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Objects,
    Terms,
)

from .combinatorial_class import CombinatorialClass, CombinatorialObject
from .exception import (
    IncorrectGeneratingFunctionError,
    InvalidOperationError,
    TaylorExpansionError,
)
from .specification_drawer import SpecificationDrawer
from .strategies import (
    AbstractStrategy,
    EmptyStrategy,
    EquivalencePathRule,
    ReverseRule,
    Rule,
    Strategy,
    StrategyPack,
    VerificationRule,
    VerificationStrategy,
)
from .strategies.rule import AbstractRule
from .utils import (
    RecursionLimit,
    maple_equations,
    pretty_print_equations,
    taylor_expand,
)

__all__ = ("CombinatorialSpecification",)


class CombinatorialSpecification(
    Generic[CombinatorialClassType, CombinatorialObjectType]
):
    """
    A combinatorial specification is a set rules of the form a -> b1, ..., bk
    where each of the bi appear exactly once on the left hand side of some
    rule.
    """

    def __init__(
        self,
        root: CombinatorialClassType,
        strategies: Iterable[
            Tuple[
                CombinatorialClassType,
                AbstractStrategy[CombinatorialClassType, CombinatorialObjectType],
            ]
        ],
        equivalence_paths: Iterable[Sequence[CombinatorialClassType]],
    ):
        self.root = root
        self.rules_dict: Dict[CombinatorialClassType, AbstractRule] = {}
        self._populate_rules_dict(strategies, equivalence_paths)
        self._remove_redundant_rules()
        self.labels: Dict[CombinatorialClassType, int] = {}
        self._label_to_tiling: Dict[int, CombinatorialClassType] = {}
        self._set_subrules()

    def _set_subrules(self) -> None:
        """Tells the subrules which children's recurrence methods it should use."""
        for rule in list(
            self.rules_dict.values()
        ):  # list as we lazily assign empty rules
            rule.set_subrecs(self.get_rule)

    def _populate_rules_dict(
        self,
        strategies: Iterable[
            Tuple[
                CombinatorialClassType,
                AbstractStrategy[CombinatorialClassType, CombinatorialObjectType],
            ]
        ],
        equivalence_paths: Iterable[Sequence[CombinatorialClassType]],
    ) -> None:
        logger.info("Creating rules.")
        equivalence_rules: Dict[
            Tuple[CombinatorialClassType, CombinatorialClassType], Rule
        ] = {}
        for comb_class, strategy in strategies:
            if isinstance(strategy, AlreadyVerified):
                continue
            rule = strategy(comb_class)
            non_empty_children = rule.non_empty_children()
            if rule.is_equivalence():
                assert isinstance(rule, Rule)
                equivalence_rules[(comb_class, non_empty_children[0])] = (
                    rule if len(rule.children) == 1 else rule.to_equivalence_rule()
                )
            elif non_empty_children or isinstance(rule, VerificationRule):
                self.rules_dict[comb_class] = rule
            else:
                raise ValueError("Non verification rule has no children.")
        self._add_equivalence_path_rules(equivalence_paths, equivalence_rules)

    def expand_verified(self) -> None:
        """
        Will expand all verified classes with respect to the strategy packs
        given by the VerificationStrategies.
        """
        verification_packs: Dict[CombinatorialClassType, StrategyPack] = {}
        for comb_class, rule in self.rules_dict.items():
            if isinstance(rule, VerificationRule):
                try:
                    verification_packs[comb_class] = rule.pack()
                except InvalidOperationError:
                    logger.info("Can't expand the rule:\n%s", rule)
        if verification_packs:
            for comb_class in verification_packs:
                self.rules_dict.pop(comb_class)
            self._expand_verified_comb_classes(verification_packs)
            self.expand_verified()
            self._set_subrules()

    def expand_comb_class(self, comb_class: Union[int, CombinatorialClassType]) -> None:
        """
        Will try to expand a particular class with respect to the strategy pack
        that the VerificationStrategy has.
        """
        rule = self.get_rule(comb_class)
        if isinstance(rule, VerificationRule):
            try:
                pack = rule.pack()
            except InvalidOperationError:
                logger.info("Can't expand the rule:\n%s", rule)
                return
            self.rules_dict.pop(rule.comb_class)
            self._expand_verified_comb_classes({rule.comb_class: pack})
        else:
            logger.info("Can't expand the rule:\n%s", rule)

    def _add_equivalence_path_rules(
        self,
        equivalence_paths: Iterable[Sequence[CombinatorialClassType]],
        equivalence_rules: Dict[
            Tuple[CombinatorialClassType, CombinatorialClassType], Rule
        ],
    ) -> None:
        logger.info("Creating equivalence path rules.")
        for eqv_path in equivalence_paths:
            if len(eqv_path) > 1:
                start = eqv_path[0]
                rules = []
                for a, b in zip(eqv_path[:-1], eqv_path[1:]):
                    try:
                        rule = equivalence_rules[(a, b)]
                    except KeyError:
                        rule = equivalence_rules[(b, a)].to_reverse_rule()
                    rules.append(rule)
                self.rules_dict[start] = EquivalencePathRule(rules)

    def _remove_redundant_rules(self) -> None:
        """
        Any redundant rules passed to the __init__ method are removed using
        this method.
        """
        rules_dict = copy(self.rules_dict)

        def prune(comb_class: CombinatorialClassType) -> None:
            try:
                rule = rules_dict.pop(comb_class)
            except KeyError:
                assert comb_class in self.rules_dict or comb_class.is_empty()
                return
            if isinstance(rule, EquivalencePathRule):
                try:
                    rule = rules_dict.pop(rule.children[0])
                except KeyError:
                    assert comb_class in self.rules_dict
                    return
            for child in rule.children:
                prune(child)

        prune(self.root)

        logger.info("Removed %s redundant rules.", len(rules_dict.values()))
        for rule in rules_dict.values():
            self.rules_dict.pop(rule.comb_class)

    def _expand_verified_comb_classes(
        self, verification_packs: Dict[CombinatorialClassType, StrategyPack]
    ) -> None:
        for comb_class, pack in verification_packs.items():
            # pylint: disable=import-outside-toplevel
            from .comb_spec_searcher import CombinatorialSpecificationSearcher

            css = CombinatorialSpecificationSearcher(
                comb_class,
                pack.add_verification(
                    AlreadyVerified(self.rules_dict), apply_first=True
                ),
            )
            logger.info(css.run_information())
            # pylint: disable=protected-access
            comb_class_rules, comb_class_eqv_paths = css._auto_search_rules()
            self._populate_rules_dict(comb_class_rules, comb_class_eqv_paths)
        self._remove_redundant_rules()

    def get_rule(
        self, comb_class: Union[int, CombinatorialClassType]
    ) -> AbstractRule[CombinatorialClassType, CombinatorialObjectType]:
        """Return the rule with comb class on the left."""
        if isinstance(comb_class, int):
            try:
                comb_class = self._label_to_tiling[comb_class]
            except KeyError as e:
                raise InvalidOperationError(
                    f"The label {comb_class} does not correspond to a tiling"
                    " in the specification."
                ) from e
        if comb_class not in self.rules_dict:
            assert comb_class.is_empty(), "rule not in the spec and not empty"
            empty_strat = EmptyStrategy()
            self.rules_dict[comb_class] = empty_strat(comb_class)
        return self.rules_dict[comb_class]

    def comb_classes(self) -> Set[CombinatorialClassType]:
        """
        Return a set containing all the combinatorial classes of the specification.
        """
        res: Set[CombinatorialClassType] = set()
        for rule in self.rules_dict.values():
            res.update(rule.children)
            res.add(rule.comb_class)
            if isinstance(rule, EquivalencePathRule):
                for parent, _ in rule.eqv_path_rules():
                    res.add(parent)
        return res

    @property
    def root_rule(
        self,
    ) -> AbstractRule[CombinatorialClassType, CombinatorialObjectType]:
        """Return the rule of the root comb class."""
        return self.rules_dict[self.root]

    def get_label(self, comb_class: CombinatorialClassType) -> int:
        """Return a unique label for the comb class."""
        res = self.labels.get(comb_class)
        if res is None:
            res = len(self.labels)
            self.labels[comb_class] = res
            self._label_to_tiling[res] = comb_class
        return res

    def get_function(self, comb_class: CombinatorialClassType) -> Function:
        """
        Return a sympy function for the comb class, using the label it is
        assigned.

        TODO: call comb_class for its parameters - 'x' is reserved for size.
        """
        x = var("x")
        extra_parameters = [var(k) for k in comb_class.extra_parameters]
        return Function("F_{}".format(self.get_label(comb_class)))(x, *extra_parameters)

    def get_equations(self) -> Iterator[Eq]:
        """
        Yield all equations on the (ordinary) generating function that the
        rules of the specification imply.
        """
        funcs: Dict[CombinatorialClass, Function] = {
            comb_class: self.get_function(comb_class)
            for comb_class, rule in self.rules_dict.items()
            if not isinstance(rule, VerificationRule)
        }
        for _, rule in sorted(
            self.rules_dict.items(), key=lambda x: self.get_label(x[0])
        ):
            try:
                if isinstance(rule, VerificationRule):
                    eq = rule.get_equation(self.get_function, funcs)
                else:
                    eq = rule.get_equation(self.get_function)
                if not isinstance(eq, bool):
                    yield eq
            except NotImplementedError:
                logger.info(
                    "can't find generating function for the rule %s -> %s. "
                    "The rule was:\n%s",
                    self.get_label(rule.comb_class),
                    tuple(self.get_label(child) for child in rule.children),
                    rule,
                )
                x = var("x")
                yield Eq(
                    self.get_function(rule.comb_class),
                    sympy.Function("NOTIMPLEMENTED")(x),
                )

    def get_initial_conditions(self, check: int = 6) -> List[Expr]:
        """
        Compute the initial conditions of the root class. It will use the
        `count_objects_of_size` method if implemented, else resort to
        the method on the `initial_conditions` method on `CombinatorialClass.
        """
        logger.info("Computing initial conditions")
        try:
            return [
                sum(
                    Number(self.count_objects_of_size(n=n, **parameters))
                    * reduce(mul, [var(k) ** val for k, val in parameters.items()], 1)
                    for parameters in self.root.possible_parameters(n)
                )
                for n in range(check + 1)
            ]
        except NotImplementedError as e:
            logger.info(
                "Reverting to generating objects from root for initial "
                "conditions due to:\nNotImplementedError: %s",
                e,
            )
        return self.root.initial_conditions(check)

    def get_genf(self, check: int = 6) -> Expr:
        """
        Return the generating function for the root comb class.

        # TODO: consider what to do if multiple variables.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        logger.info("Computing initial conditions")
        initial_conditions = self.get_initial_conditions(check)
        logger.info(pretty_print_equations(root_func, initial_conditions, eqs))
        logger.info("Solving...")
        solutions = solve(
            eqs,
            tuple([eq.lhs for eq in eqs]),
            dict=True,
            cubics=False,
            quartics=False,
            quintics=False,
        )
        for solution in solutions:
            genf = solution[root_func]
            logger.info("Checking initial conditions for: %s", genf)
            try:
                expansion = taylor_expand(genf, check)
            except TaylorExpansionError:
                continue
            if expansion == initial_conditions:
                return sympy.simplify(genf)
        raise IncorrectGeneratingFunctionError(
            "Failed to compute the generating function for the specification."
        )

    def get_maple_equations(self, check: int = 6) -> str:
        """
        Convert the systems of equations to version that can be copy pasted to maple.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        initial_conditions = self.get_initial_conditions(check)
        maple_eqs = maple_equations(root_func, initial_conditions, eqs)
        return maple_eqs

    def get_pretty_equations(self, check: int = 6) -> str:
        """
        Convert the systems of equations to a more readable format.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        initial_conditions = self.get_initial_conditions(check)
        return pretty_print_equations(root_func, initial_conditions, eqs)

    def count_objects_of_size(self, n: int, **parameters) -> int:
        """
        Return the number of objects with the given parameters.
        Note, 'n' is reserved for the size of the object.
        """
        limit = n * self.number_of_rules()
        with RecursionLimit(limit):
            return self.root_rule.count_objects_of_size(n, **parameters)

    def get_terms(self, n: int) -> Terms:
        """
        Return the terms for given n.
        """
        return self.root_rule.get_terms(n)

    def get_objects(self, n: int) -> Objects:
        """
        Return the objects for given n.
        """
        return self.root_rule.get_objects(n)

    def generate_objects_of_size(
        self, n: int, **parameters
    ) -> Iterator[CombinatorialObjectType]:
        """
        Return the objects with the given parameters.
        Note, 'n' is reserved for the size of the object.
        """
        for obj in self.root_rule.generate_objects_of_size(n, **parameters):
            yield obj

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        """
        Return a uniformly random object of the given size. This is done using
        the "recursive" method.
        """
        limit = n * self.number_of_rules()
        with RecursionLimit(limit):
            return self.root_rule.random_sample_object_of_size(n, **parameters)

    def number_of_rules(self) -> int:
        return len(self.rules_dict)

    def sanity_check(self, length: int = 5) -> bool:
        """
        Sanity check the specification to the given length.

        Raise an SanityCheckFailure error if it fails.
        """
        return all(
            all(rule.sanity_check(n) for rule in self.rules_dict.values())
            for n in range(length + 1)
        )

    def show(
        self, levels_shown: int = 0, levels_expand: int = 0, verbose: bool = False
    ) -> None:
        """
        Displays a tree representing this object in the web browser
        OTHER INPUT:
            - 'levels_shown': number of levels displayed at the start.
            If 0 then the whole tree is displayed
            - 'levels_expand': number of levels displayed after expanding a node.
            If 0 then the rest of the tree is displayed
            - 'verbose': displays formal step below the node
            and json representation in tooltips
        """
        sd = SpecificationDrawer(
            self,
            levels_shown=levels_shown,
            levels_expand=levels_expand,
            verbose=verbose,
        )
        sd.show()

    def share(
        self, levels_shown: int = 0, levels_expand: int = 0, verbose: bool = False
    ) -> None:
        """
        Uploads the html representation of the specification to a sharing website and
        display the link to the file.

        INPUT:
            - 'levels_shown': number of levels displayed at the start.
            If 0 then the whole tree is displayed
            - 'levels_expand': number of levels displayed after expanding a node.
            If 0 then the rest of the tree is displayed
            - 'verbose': displays formal step below the node
            and json representation in tooltips
        """
        sd = SpecificationDrawer(
            self,
            levels_shown=levels_shown,
            levels_expand=levels_expand,
            verbose=verbose,
        )
        sd.share()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CombinatorialSpecification):
            return NotImplemented
        return bool(self.root == other.root and self.rules_dict == other.rules_dict)

    def __str__(self) -> str:
        res = "A combinatorial specification with {} rules.".format(
            self.number_of_rules()
        )
        rules_dict = copy(self.rules_dict)

        def update_res(comb_class: CombinatorialClassType, res: str):
            try:
                rule = rules_dict.pop(comb_class)
            except KeyError:
                assert comb_class in self.rules_dict or comb_class.is_empty()
                return res
            if isinstance(rule, EquivalencePathRule):
                child_label = self.get_label(rule.comb_class)
                child_eqv_label = self.get_label(rule.children[0])
                labels = "{} = {}\n".format(child_label, child_eqv_label)
                res += "\n{}\n".format("-" * (len(labels) - 1))
                res += labels
                res += str(rule)
                try:
                    rule = rules_dict.pop(rule.children[0])
                except KeyError:
                    assert comb_class in self.rules_dict
                    return res
            start_label = self.get_label(rule.comb_class)
            end_labels = tuple(self.get_label(c) for c in rule.children)
            labels = "{} -> {}\n".format(start_label, end_labels)
            res += "\n{}\n".format("-" * (len(labels) - 1))
            res += labels
            res += str(rule)
            for child in rule.children:
                res = update_res(child, res)
            return res

        res = update_res(self.root, res)
        assert not rules_dict
        return res + "\n"

    def equations_string(self) -> str:
        """
        Return a convenient to read string version of the equations returned by
        'get_equations' method.
        """
        res = ""
        for eq in sorted(
            self.get_equations(),
            key=lambda eq: int(str(eq.lhs).split("_")[1].split("(")[0]),
        ):
            res += "\n{} = {}".format(eq.lhs, eq.rhs)
        return res

    # JSON methods

    def to_jsonable(self) -> dict:
        """Return a JSON serializable dictionary for the specification."""
        rules = [
            rule
            for rule in self.rules_dict.values()
            if not isinstance(rule, EquivalencePathRule)
        ]
        eqv_paths = []
        for rule in self.rules_dict.values():
            if isinstance(rule, EquivalencePathRule):
                eqv_path = []
                for c, r in rule.eqv_path_rules():
                    rules.append(r)
                    eqv_path.append(c.to_jsonable())
                eqv_path.append(rule.children[0].to_jsonable())
                eqv_paths.append(eqv_path)
        strategies = [
            (rule.children[0].to_jsonable(), rule.strategy.to_jsonable())
            if isinstance(rule, ReverseRule)
            else (rule.comb_class.to_jsonable(), rule.strategy.to_jsonable())
            for rule in rules
        ]
        return {
            "root": self.root.to_jsonable(),
            "strategies": strategies,
            "eqv_paths": eqv_paths,
        }

    @classmethod
    def from_dict(cls, d):
        """
        Return the specification with the dictionary outputter by the
        'to_jsonable' method
        """
        strategies = []
        for comb_class_dict, strategy_dict in d["strategies"]:
            comb_class = CombinatorialClass.from_dict(comb_class_dict)
            strategy = Strategy.from_dict(strategy_dict)
            assert isinstance(strategy, (Strategy, VerificationStrategy))
            strategies.append((comb_class, strategy))
        return cls(
            root=CombinatorialClass.from_dict(d["root"]),
            strategies=strategies,
            equivalence_paths=[
                [CombinatorialClass.from_dict(class_dict) for class_dict in eqv_path]
                for eqv_path in d["eqv_paths"]
            ],
        )


class AlreadyVerified(VerificationStrategy[CombinatorialClass, CombinatorialObject]):
    def __init__(self, verfied_comb_classes: Iterable[CombinatorialClass]):
        self.verified_classes = frozenset(verfied_comb_classes)
        super().__init__()

    def verified(self, comb_class: CombinatorialClass) -> bool:
        return comb_class in self.verified_classes

    @staticmethod
    def formal_step() -> str:
        return "already verified"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["comb_classes"] = [c.to_jsonable() for c in self.verified_classes]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AlreadyVerified":
        return cls([CombinatorialClass.from_dict(c) for c in d["comb_classes"]])



"""Some custom errors."""


class ExceededMaxtimeError(Exception):
    """Ran out of time."""


class IncorrectGeneratingFunctionError(Exception):
    """The incorrect generating function was found."""


class InsaneTreeError(Exception):
    """A sanity check failed."""


class InvalidOperationError(Exception):
    """The operation used doesn't apply."""


class ObjectMappingError(Exception):
    """A problem in the bijection."""


class SanityCheckFailure(Exception):
    """Failed a sanity check."""


class SpecificationNotFound(Exception):
    """A specification was not found."""


class StrategyDoesNotApply(Exception):
    """A strategy does not apply to the combinatorial class."""


class NoMoreClassesToExpandError(Exception):
    """The class queue has run out of classes to try and expand."""


class TaylorExpansionError(Exception):
    """An error while taylor expanding."""



from typing import (
    TYPE_CHECKING,
    Callable,
    Counter,
    DefaultDict,
    Dict,
    List,
    NamedTuple,
    Set,
    Tuple,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from comb_spec_searcher.combinatorial_class import CombinatorialClass  # noqa: F401
    from comb_spec_searcher.combinatorial_class import CombinatorialObject  # noqa: F401
    from comb_spec_searcher.strategies.strategy import AbstractStrategy, StrategyFactory

__all__ = [
    "CombinatorialObjectType",
    "CombinatorialClassType",
    "SpecificationClassesAndStrats",
]

CombinatorialObjectType = TypeVar(
    "CombinatorialObjectType", bound="CombinatorialObject"
)

CombinatorialClassType = TypeVar("CombinatorialClassType", bound="CombinatorialClass")


SpecificationClassesAndStrats = Tuple[
    List[Tuple[CombinatorialClassType, "AbstractStrategy"]],
    List[List[CombinatorialClassType]],
]
ClassKey = Union[bytes, CombinatorialClassType]
Key = Union["CombinatorialClassType", int]

CSSstrategy = Union["AbstractStrategy", "StrategyFactory"]
RulesDict = Dict[int, Set[Tuple[int, ...]]]


class WorkPacket(NamedTuple):
    label: int
    strategies: Tuple[CSSstrategy, ...]
    inferral: bool


SpecificationLabelsAndStrats = Tuple[
    List[Tuple[int, "AbstractStrategy"]], List[List[int]]
]
RuleKey = Tuple[int, Tuple[int, ...]]

# From constructor
Parameters = Tuple[int, ...]
ParametersMap = Callable[[Parameters], Parameters]
RelianceProfile = Tuple[Dict[str, Tuple[int, ...]], ...]
Objects = DefaultDict[Parameters, List[CombinatorialObjectType]]
ObjectsCache = List[Objects]
Terms = Counter[Parameters]  # all terms for a fixed n
TermsCache = List[Terms]  # index n contains terms for n
SubObjects = Tuple[Callable[[int], Objects], ...]
SubRecs = Tuple[Callable[..., int], ...]
SubSamplers = Tuple[Callable[..., CombinatorialObjectType], ...]
SubTerms = Tuple[Callable[[int], Terms], ...]
